"""Test container commands."""

from __future__ import annotations

import platform
from collections import defaultdict

import pytest

from meltano.core.container.container_spec import ContainerSpec


class TestContainerService:
    @pytest.mark.parametrize(
        ("spec", "payload"),
        (
            (
                ContainerSpec(
                    "lightdash/lightdash",
                    ports={"8080": "8000"},
                ),
                {
                    "Cmd": None,
                    "Entrypoint": None,
                    "Image": "lightdash/lightdash",
                    "Env": [],
                    "ExposedPorts": {"8000": {}},
                    "HostConfig": {
                        "PortBindings": defaultdict(
                            list,
                            {
                                "8000": [
                                    {
                                        "HostPort": "8080",
                                        "HostIP": "0.0.0.0",  # noqa: S104v
                                    },
                                ],
                            },
                        ),
                        "Binds": [],
                    },
                },
            ),
            (
                ContainerSpec(
                    "lightdash/lightdash",
                    entrypoint="bash",
                ),
                {
                    "Cmd": None,
                    "Entrypoint": "bash",
                    "Image": "lightdash/lightdash",
                    "Env": [],
                    "ExposedPorts": {},
                    "HostConfig": {
                        "PortBindings": defaultdict(list, {}),
                        "Binds": [],
                    },
                },
            ),
        ),
        ids=["port-mapping", "custom-entrypoint"],
    )
    @pytest.mark.asyncio()
    async def test_docker_config(self, spec: ContainerSpec, payload: dict) -> None:
        """Check Docker container config from container spec."""
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )
        config = spec.get_docker_config()
        assert config == payload



"""Create the EmbedToken table.

Revision ID: ceb00d7ff3bd
Revises: 990c0665f3ce
Create Date: 2020-01-27 09:59:36.427138

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "ceb00d7ff3bd"
down_revision = "990c0665f3ce"
branch_labels = None
depends_on = None


def upgrade() -> None:
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)
    max_string_length = max_string_length_for_dialect(dialect_name)

    op.create_table(
        "embed_tokens",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("token", sa.String(64), unique=True),
        sa.Column("resource_id", sa.String(max_string_length), nullable=False),
        sa.Column("created_at", datetime_type),
    )


def downgrade() -> None:
    op.drop_table("embed_tokens")



from __future__ import annotations

import typing as t

import click
import mock
import pytest
from requests import HTTPError, Response
from requests.adapters import BaseAdapter

from meltano.cli import cli
from meltano.cli.hub import hub
from meltano.core.hub.client import HubConnectionError, HubPluginVariantNotFoundError
from meltano.core.plugin.base import PluginType, Variant
from meltano.core.plugin.error import PluginNotFoundError

if t.TYPE_CHECKING:
    from collections import Counter

    from meltano.core.project import Project


class TestMeltanoHubService:
    def test_find_definition_specified_variant(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        definition = project.hub_service.find_definition(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant_name="meltano",
        )
        assert definition.name == "tap-mock"
        assert definition.variants[0].name == "meltano"

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/extractors/tap-mock--meltano"] == 1

    def test_find_definition_default_variant(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        definition = project.hub_service.find_definition(
            PluginType.EXTRACTORS,
            "tap-mock",
        )
        assert definition.name == "tap-mock"
        assert definition.variants[0].name == "meltano"

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/extractors/tap-mock--meltano"] == 1

    def test_find_definition_original_variant(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        definition = project.hub_service.find_definition(
            PluginType.EXTRACTORS,
            "tap-mock",
            variant_name=Variant.ORIGINAL_NAME,
        )
        assert definition.name == "tap-mock"
        assert definition.variants[0].name == "meltano"

        assert hub_request_counter["/extractors/index"] == 1
        assert hub_request_counter["/extractors/tap-mock--meltano"] == 1

    def test_definition_not_found(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        with pytest.raises(PluginNotFoundError):
            project.hub_service.find_definition(PluginType.EXTRACTORS, "tap-not-found")

        assert hub_request_counter["/extractors/index"] == 1
        assert len(hub_request_counter) == 1

    def test_variant_not_found(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        with pytest.raises(HubPluginVariantNotFoundError):
            project.hub_service.find_definition(
                PluginType.EXTRACTORS,
                "tap-mock",
                "not-found",
            )

        assert hub_request_counter["/extractors/index"] == 1
        assert len(hub_request_counter) == 1

    def test_get_plugins_of_type(
        self,
        project: Project,
        hub_request_counter: Counter,
    ) -> None:
        extractors = project.hub_service.get_plugins_of_type(PluginType.EXTRACTORS)
        assert len(extractors) == 9
        assert len(extractors["tap-mock"].variants) == 2
        assert extractors["tap-mock"].variant_labels == [
            "meltano (default)",
            "singer-io",
        ]
        assert hub_request_counter["/extractors/index"] == 1

    def test_hub_auth(self, project) -> None:
        project.settings.set("hub_url_auth", "Bearer s3cr3t")
        assert project.hub_service.session.headers["Authorization"] == "Bearer s3cr3t"

    def test_server_error(self, project: Project) -> None:
        with pytest.raises(
            HubConnectionError,
            match="Could not connect to Meltano Hub. 500 Server Error",
        ) as exc_info:
            project.hub_service.find_definition(
                PluginType.EXTRACTORS,
                "this-returns-500",
            )

        assert isinstance(exc_info.value.__cause__, HTTPError)
        assert isinstance(exc_info.value.__cause__.response, Response)
        assert exc_info.value.__cause__.response.status_code == 500
        assert exc_info.value.__cause__.response.json() == {"error": "Server error"}
        assert exc_info.value.__cause__.response.url == (
            "https://hub.meltano.com/meltano/api/v1/plugins/extractors"
            "/this-returns-500--original"
        )

    def test_request_headers(self, project: Project) -> None:
        with mock.patch("click.get_current_context") as get_context:
            get_context.return_value = click.Context(
                hub,
                info_name="hub",
                parent=click.Context(cli, info_name="meltano"),
            )
            request = project.hub_service._build_request("GET", "https://example.com")
            assert request.headers["X-Meltano-Command"] == "meltano hub"

        with mock.patch("click.get_current_context") as get_context:
            get_context.return_value = None
            request = project.hub_service._build_request("GET", "https://example.com")
            assert "X-Meltano-Command" not in request.headers

    def test_custom_ca(self, project, monkeypatch) -> None:
        send_kwargs = {}

        class _Adapter(BaseAdapter):
            def send(
                self,
                request,  # noqa: ARG002
                **kwargs,
            ):
                nonlocal send_kwargs
                send_kwargs = kwargs

                response = Response()
                response._content = b'{"name": "tap-mock", "namespace": "tap_mock"}'
                response.status_code = 200
                return response

        mock_url = "hub://meltano"
        hub = project.hub_service
        hub.session.mount(mock_url, _Adapter())

        monkeypatch.setenv("REQUESTS_CA_BUNDLE", "/path/to/ca.pem")
        hub._get(mock_url)
        assert send_kwargs["verify"] == "/path/to/ca.pem"

    def test_custom_proxy(self, project, monkeypatch) -> None:
        send_kwargs = {}

        class _Adapter(BaseAdapter):
            def send(
                self,
                request,  # noqa: ARG002
                **kwargs,
            ):
                nonlocal send_kwargs
                send_kwargs = kwargs

                response = Response()
                response._content = b'{"name": "tap-mock", "namespace": "tap_mock"}'
                response.status_code = 200
                return response

        mock_url = "hub://meltano"
        hub = project.hub_service
        hub.session.mount(mock_url, _Adapter())

        monkeypatch.setenv("HTTPS_PROXY", "https://www.example.com:3128/")
        hub._get(mock_url)
        assert send_kwargs["proxies"] == {"https": "https://www.example.com:3128/"}



"""StateStoreManager for S3 cloud storage backend."""

from __future__ import annotations

import typing as t
from contextlib import contextmanager
from functools import cached_property

from meltano.core.state_store.filesystem import (
    CloudStateStoreManager,
    InvalidStateBackendConfigurationException,
)

if t.TYPE_CHECKING:
    from collections.abc import Iterator

    from mypy_boto3_s3 import S3Client

BOTO_INSTALLED = True

try:
    import boto3
except ImportError:
    BOTO_INSTALLED = False


class MissingBoto3Error(Exception):
    """Raised when boto3 is required but not installed."""

    def __init__(self) -> None:
        """Initialize a MissingBoto3Error."""
        super().__init__(
            "boto3 required but not installed. Install meltano[s3] to use S3 as a state backend.",  # noqa: E501
        )


@contextmanager
def requires_boto3():  # noqa: ANN201
    """Raise MissingBoto3Error if boto3 is required but missing in context.

    Raises:
        MissingBoto3Error: if boto3 is not installed.

    Yields:
        None
    """
    if not BOTO_INSTALLED:
        raise MissingBoto3Error
    yield


class S3StateStoreManager(CloudStateStoreManager):
    """State backend for S3."""

    label: str = "AWS S3"

    def __init__(
        self,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        bucket: str | None = None,
        prefix: str | None = None,
        endpoint_url: str | None = None,
        **kwargs,  # noqa: ANN003
    ):
        """Initialize the BaseFilesystemStateStoreManager.

        Args:
            aws_access_key_id: AWS access key ID used to authenticate
            aws_secret_access_key: AWS secret access key used to authenticate
            bucket: the bucket to store state in
            prefix: the prefix to store state at
            endpoint_url: the endpoint url for S3
            kwargs: additional keyword args to pass to parent
        """
        super().__init__(**kwargs)
        self.aws_access_key_id = aws_access_key_id or self.parsed.username
        self.aws_secret_access_key = aws_secret_access_key or self.parsed.password

        # TODO: Make this type-safe. Maybe use https://github.com/fsspec/universal_pathlib?
        self.bucket: str = bucket or self.parsed.hostname  # type: ignore[assignment]
        self.prefix = prefix or self.parsed.path
        self.endpoint_url = endpoint_url

    @staticmethod
    def is_file_not_found_error(err: Exception) -> bool:
        """Check if err is equivalent to file not being found.

        Args:
            err: the err to check

        Returns:
            True if error represents file not being found, else False
        """
        return (
            isinstance(err, OSError)
            and "NoSuchKey" in err.args[0]
            or (isinstance(err, KeyError) and "ActualObjectSize" in err.args[0])
        )

    @cached_property
    def client(self) -> S3Client:
        """Get an authenticated boto3.Client.

        Returns:
            A boto3.Client.

        Raises:
            InvalidStateBackendConfigurationException: when configured AWS
                settings are invalid.
        """
        with requires_boto3():
            if self.aws_secret_access_key and self.aws_access_key_id:
                session = boto3.Session(
                    aws_access_key_id=self.aws_access_key_id,
                    aws_secret_access_key=self.aws_secret_access_key,
                )
                return session.client("s3", endpoint_url=self.endpoint_url)
            if self.aws_secret_access_key:
                raise InvalidStateBackendConfigurationException(
                    "AWS secret access key configured, but not AWS access key ID.",  # noqa: EM101
                )
            if self.aws_access_key_id:
                raise InvalidStateBackendConfigurationException(
                    "AWS access key ID configured, but no AWS secret access key.",  # noqa: EM101
                )
            session = boto3.Session()
            return session.client("s3")

    def delete(self, file_path: str) -> None:
        """Delete the file/blob at the given path.

        Args:
            file_path: the path to delete.
        """
        self.client.delete_objects(
            Bucket=self.bucket,
            Delete={"Objects": [{"Key": file_path}]},
        )

    def list_all_files(self, *, with_prefix: bool = True) -> Iterator[str]:
        """List all files in the backend.

        Args:
            with_prefix: Whether to include the prefix in the lookup.

        Yields:
            The path to each file in the backend.
        """
        kwargs: dict[str, t.Any] = {"Bucket": self.bucket}
        if with_prefix:
            kwargs["Prefix"] = self.prefix

        for state_obj in self.client.list_objects_v2(**kwargs).get("Contents", []):
            yield state_obj["Key"]

    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file from one path to another.

        Args:
            src: the source path
            dst: the destination path
        """
        self.client.copy_object(
            Bucket=self.bucket,
            CopySource={"Bucket": self.bucket, "Key": src},
            Key=dst,
        )



from __future__ import annotations  # noqa: D100


class NameEq:  # noqa: D101
    def __eq__(self, other):  # noqa: ANN001, ANN204, D105
        return self is other or self.name == other.name

    def __hash__(self):  # noqa: ANN204, D105
        return hash(self.name)



from __future__ import annotations

import inspect
import json
import platform
import sys
import typing as t
import uuid
import warnings
from pathlib import Path
from platform import python_version_tuple

import pytest
from jsonschema import ValidationError, validate

from meltano.core.tracking import __file__ as tracking_module_path
from meltano.core.tracking.contexts import ExceptionContext
from meltano.core.utils import hash_sha256

THIS_FILE_BASENAME = Path(__file__).name

with (
    Path(tracking_module_path).parent
    / "iglu-client-embedded"
    / "schemas"
    / "com.meltano"
    / "exception_context"
    / "jsonschema"
    / "1-0-0"
).open() as exception_context_schema_file:
    EXCEPTION_CONTEXT_SCHEMA = json.load(exception_context_schema_file)


class CustomException(Exception):
    """A custom exception type to be used in `test_complex_exception_context`."""


def is_valid_exception_context(instance: dict[str, t.Any]) -> bool:
    try:
        with warnings.catch_warnings():
            # Ignore the misleading warning thrown by `jsonschema`:
            #     The metaschema specified by `$schema` was not found. Using
            #     the latest draft to validate, but this will raise an error
            #     in the future.
            # This is a bug in `jsonschema`, as our value for `$schema` is fine.
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            validate(instance, EXCEPTION_CONTEXT_SCHEMA)
    except ValidationError:
        return False
    return True


def test_null_exception_context() -> None:
    ctx = ExceptionContext()
    assert isinstance(ctx.data, dict)
    assert isinstance(ctx.data["context_uuid"], str)
    uuid.UUID(ctx.data["context_uuid"], version=4)
    assert ctx.data["exception"] is None


def test_simple_exception_context() -> None:
    msg = "The error message"

    ex = ValueError(msg)
    try:
        raise ex
    except Exception:
        ctx = ExceptionContext()

    assert is_valid_exception_context(ctx.data)

    ex_data = ctx.data["exception"]

    assert ex_data["type"] == "ValueError"
    assert (
        ex_data["str_hash"]
        == "84e45725cb2d98b23d365ff46d8aae35a23c6d91fbed6c0b19ad2ddb72e66f01"
        == hash_sha256(str(ex))
        == hash_sha256(msg)
    )
    assert (
        ex_data["repr_hash"]
        == "544806225e70d8ae22a4826a02b0be4320a71c1d3b5f0aa4630828815bee76b1"
        == hash_sha256(repr(ex))
        == hash_sha256(f"ValueError({msg!r})")
    )

    assert ex_data["cause"] is None
    assert ex_data["context"] is None

    tb_data = ex_data["traceback"]
    assert len(tb_data) == 1
    for key in ("file", "line_number"):
        assert key in tb_data[0]


def test_complex_exception_context() -> None:
    if platform.system() == "Windows":
        pytest.xfail("Fails on Windows: https://github.com/meltano/meltano/issues/3444")

    line_nums: list[int] = []
    file_not_found_error = None

    def _function_to_deepen_traceback() -> None:
        try:
            line_nums.append(1 + inspect.currentframe().f_lineno)
            Path("/tmp/fake/path/will/not/resolve").resolve(strict=True)  # noqa: S108
        except Exception as ex:
            nonlocal file_not_found_error
            file_not_found_error = ex
            line_nums.append(1 + inspect.currentframe().f_lineno)
            raise ValueError("that path was a bad value") from ex  # noqa: EM101

    try:
        try:
            line_nums.append(1 + inspect.currentframe().f_lineno)
            _function_to_deepen_traceback()
        except Exception:
            line_nums.append(1 + inspect.currentframe().f_lineno)
            raise CustomException from None
    except Exception:
        ctx = ExceptionContext()

    assert is_valid_exception_context(ctx.data)

    major, minor, _ = python_version_tuple()

    cause = ctx.data["exception"]["context"].pop("cause")
    context = ctx.data["exception"]["context"].pop("context")

    assert cause == context
    assert cause["type"] == "FileNotFoundError"
    assert cause["str_hash"] == hash_sha256(str(file_not_found_error))
    assert cause["repr_hash"] == hash_sha256(repr(file_not_found_error))
    assert cause["traceback"][0] == {
        "file": f".../{THIS_FILE_BASENAME}",
        "line_number": line_nums[1],
    }

    if sys.version_info < (3, 13):
        pathlib_loc = "pathlib.py"
    else:
        # https://github.com/python/cpython/pull/118582
        pathlib_loc = "pathlib/_local.py"

    assert cause["traceback"][1]["file"] == f"lib/python{major}.{minor}/{pathlib_loc}"
    assert cause["cause"] is None
    assert cause["context"] is None

    assert ctx.data == {
        "context_uuid": ctx.data["context_uuid"],
        "exception": {
            "type": "CustomException",
            "str_hash": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",  # noqa: E501
            "repr_hash": "ad9443d77d731da456747bd47282a51afe86be7058533f44dcc979320ad62c73",  # noqa: E501
            "traceback": [
                {
                    "file": f".../{THIS_FILE_BASENAME}",
                    "line_number": line_nums[3],
                },
            ],
            "cause": None,
            "context": {
                "type": "ValueError",
                "str_hash": "1009263b7f48917b8f0edafcfc8a06d22156122fbcbfbb7c9139f420b8472e0c",  # noqa: E501
                "repr_hash": "0015450e35aed13f4802973752ee45d02c8f8eaa5d57417962986f4b8ef1bf88",  # noqa: E501
                "traceback": [
                    {
                        "file": f".../{THIS_FILE_BASENAME}",
                        "line_number": line_nums[0],
                    },
                    {
                        "file": f".../{THIS_FILE_BASENAME}",
                        "line_number": line_nums[2],
                    },
                ],
            },
        },
    }



"""Add plugins to the project."""

from __future__ import annotations

import enum
import typing as t

from meltano.core.plugin import BasePlugin, PluginType, Variant
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.project_plugins_service import (
    DefinitionSource,
    PluginAlreadyAddedException,
)

if t.TYPE_CHECKING:
    from meltano.core.project import Project


class PluginAddedReason(str, enum.Enum):
    """The reason why a plugin was added to the project."""

    #: The plugin was added by the user.
    USER_REQUEST = "user_request"

    #: The plugin was added because it is related to another plugin.
    RELATED = "related"

    #: The plugin was added because it is required by another plugin.
    REQUIRED = "required"


class MissingPluginException(Exception):
    """Raised when a plugin is not found."""


class ProjectAddService:
    """Project Add Service."""

    def __init__(self, project: Project):
        """Create a new Project Add Service.

        Args:
            project: The project to add plugins to.
        """
        self.project = project

    def add(
        self,
        plugin_type: PluginType,
        plugin_name: str,
        *,
        lock: bool = True,
        update: bool = False,
        **attrs: t.Any,
    ) -> ProjectPlugin:
        """Add a new plugin to the project.

        Args:
            plugin_type: The type of the plugin to add.
            plugin_name: The name of the plugin to add.
            lock: Whether to generate a lockfile for the plugin.
            update: Whether to update the plugin.
            attrs: Additional attributes to add to the plugin.

        Returns:
            The added plugin.
        """
        plugin = ProjectPlugin(
            plugin_type,
            plugin_name,
            **attrs,
            default_variant=Variant.DEFAULT_NAME,
        )

        with self.project.plugins.use_preferred_source(DefinitionSource.ANY):
            self.project.plugins.ensure_parent(plugin)

            # If we are inheriting from a base plugin definition,
            # repeat the variant and pip_url in meltano.yml
            parent = plugin.parent
            if isinstance(parent, BasePlugin):
                plugin.variant = parent.variant
                plugin.pip_url = parent.pip_url

            if update:
                plugin, _outdated = self.project.plugins.update_plugin(
                    plugin,
                    keep_config=True,
                )
            else:
                plugin = self.project.plugins.add_to_file(plugin)

            if lock and not plugin.is_custom():
                self.project.plugins.lock_service.save(
                    plugin,
                    exists_ok=update or plugin.inherit_from is not None,
                )

            return plugin

    def add_required(
        self,
        plugin: ProjectPlugin,
        *,
        lock: bool = True,
    ) -> list[ProjectPlugin]:
        """Add all required plugins to the project.

        Args:
            plugin: The plugin to get requirements from.
            lock: Whether to generate a lockfile for the plugin.

        Returns:
            The added plugins.
        """
        added_plugins: list[ProjectPlugin] = []
        for plugin_type, plugins in plugin.all_requires.items():
            for plugin_req in plugins:
                try:
                    plugin = self.add(
                        plugin_type,
                        plugin_req.name,
                        variant=plugin_req.variant,
                        lock=lock,
                    )
                except PluginAlreadyAddedException:
                    continue

                added_plugins.append(plugin)

        added_plugins_with_required: list[ProjectPlugin] = []
        for added in added_plugins:
            added_plugins_with_required.extend([added, *self.add_required(added)])

        return added_plugins_with_required



from __future__ import annotations  # noqa: D100

from alembic import context
from sqlalchemy import engine_from_config, pool

config = context.config


def run_migrations_offline():  # noqa: ANN201
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():  # noqa: ANN201
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    if connection := config.attributes.get("connection"):
        _run_migrations_online(connection)
        return

    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as new_connection:
        _run_migrations_online(new_connection)


def _run_migrations_online(connection):  # noqa: ANN001, ANN202
    context.configure(connection=connection)

    with context.begin_transaction():
        context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Shared Pytest fixture utilities."""

from __future__ import annotations

import logging
import os
import typing as t
from contextlib import contextmanager
from pathlib import Path

from meltano.core.project import Project
from meltano.core.project_init_service import ProjectInitService


@contextmanager
def cd(path: Path) -> t.Generator[Path, None, None]:
    prev_dir = Path.cwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(prev_dir)


@contextmanager
def tmp_project(
    name: str,
    source: Path,
    compatible_copy_tree,
) -> t.Generator[Project, None, None]:
    project_init_service = ProjectInitService(name)
    blank_project = project_init_service.init()
    logging.debug(f"Created new project at {blank_project.root}")  # noqa: G004, TID251
    blank_project.meltanofile.unlink()
    compatible_copy_tree(source, blank_project.root)
    Project._default = None
    project = Project(blank_project.root)
    with cd(project.root):
        try:
            project.refresh()
            yield project
        finally:
            Project.deactivate()
            logging.debug(f"Cleaned project at {project.root}")  # noqa: G004, TID251



from __future__ import annotations

import pytest

from meltano.cli import cli
from meltano.core.error import ProjectNotFound
from meltano.core.project import Project
from meltano.core.project_init_service import ProjectInitServiceError


class TestCliInit:
    def test_init(self, cli_runner, tmp_path_factory, pushd) -> None:
        new_project_root = tmp_path_factory.mktemp("new_meltano_root")
        pushd(new_project_root)

        Project.deactivate()
        with pytest.raises(ProjectNotFound):
            Project.find()

        # Create a project with the CLI
        cli_runner.invoke(cli, ["init", "test_project", "--no-usage-stats"])

        pushd("test_project")
        project = Project.find()
        Project.deactivate()

        files = (
            project.root.joinpath(file).resolve()
            for file in ("meltano.yml", "README.md", ".gitignore", "requirements.txt")
        )

        meltano_dirs = (
            project.root.joinpath(meltano_dir)
            for meltano_dir in (
                ".meltano",
                "extract",
                "load",
                "transform",
                "analyze",
                "notebook",
                "orchestrate",
            )
        )

        for file in files:
            assert file.is_file()

        for meltano_dir in meltano_dirs:
            assert meltano_dir.is_dir()

        meltano_yml = project.root_dir("meltano.yml").read_text()
        assert "send_anonymous_usage_stats: false" in meltano_yml
        assert "project_id:" in meltano_yml

    def test_init_existing_empty_directory(
        self,
        cli_runner,
        tmp_path_factory,
        pushd,
    ) -> None:
        new_project_root = tmp_path_factory.mktemp("new_meltano_root")
        pushd(new_project_root)

        # create project in empty current working directory
        result = cli_runner.invoke(cli, ["init", ".", "--no_usage_stats"])
        assert "Creating project files...\n  ./" not in result.output
        assert "cd ." not in result.output

        Project.deactivate()

        new_project_root = tmp_path_factory.mktemp("new_meltano_root")
        pushd(new_project_root)

        # create project in another empty directory
        project_dir = new_project_root.joinpath("test_project")
        project_dir.mkdir()
        assert project_dir.exists()

        result = cli_runner.invoke(cli, ["init", "test_project", "--no_usage-stats"])
        assert "Creating project files...\n  test_project/" in result.output
        assert "cd test_project" in result.output

        Project.deactivate()

    def test_init_existing_meltano_yml(
        self,
        cli_runner,
        tmp_path_factory,
        pushd,
    ) -> None:
        new_project_root = tmp_path_factory.mktemp("new_meltano_root")
        new_project_root.joinpath("meltano.yml").touch()
        new_project_root.joinpath("README.md").touch()
        pushd(new_project_root)

        # create project in non-empty current working directory
        result = cli_runner.invoke(cli, ["init", ".", "--no_usage_stats"])
        assert result.exit_code == 1
        assert isinstance(result.exception, ProjectInitServiceError)
        assert "A `meltano.yml` file already exists" in result.exception.args[0]

        # forcefully create project in non-empty current working directory
        result = cli_runner.invoke(cli, ["init", ".", "--no_usage_stats", "--force"])
        assert result.exit_code == 0
        assert "Creating project files..." in result.output
        assert "meltano.yml" in result.output
        assert "README.md (skipped)" in result.output

        Project.deactivate()



"""Environment management in CLI."""

from __future__ import annotations

import typing as t

import click

from meltano.cli.params import pass_project
from meltano.cli.utils import InstrumentedGroup, PartialInstrumentedCmd
from meltano.core.environment_service import EnvironmentService
from meltano.core.tracking.contexts import CliEvent

if t.TYPE_CHECKING:
    from meltano.core.project import Project

ENVIRONMENT_SERVICE_KEY = "environment_service"


@click.group(
    cls=InstrumentedGroup,
    name="environment",
    short_help="Manage environments.",
)
@click.pass_context
@pass_project(migrate=True)
def meltano_environment(project: Project, ctx: click.Context) -> None:
    """Manage Environments.

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#environment
    """  # noqa: D301
    ctx.obj[ENVIRONMENT_SERVICE_KEY] = EnvironmentService(project)


@meltano_environment.command(cls=PartialInstrumentedCmd)
@click.argument("name")
@click.pass_context
def add(ctx: click.Context, name: str) -> None:
    """Add a new environment."""
    tracker = ctx.obj["tracker"]
    environment_service: EnvironmentService = ctx.obj[ENVIRONMENT_SERVICE_KEY]
    try:
        environment = environment_service.add(name)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    click.echo(f"Created new environment '{environment.name}'")
    tracker.track_command_event(CliEvent.completed)


@meltano_environment.command(cls=PartialInstrumentedCmd)
@click.argument("name")
@click.pass_context
def remove(ctx: click.Context, name: str) -> None:
    """Remove an environment."""
    tracker = ctx.obj["tracker"]
    environment_service: EnvironmentService = ctx.obj[ENVIRONMENT_SERVICE_KEY]
    try:
        environment_name = environment_service.remove(name)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    click.echo(f"Removed environment '{environment_name}'")
    tracker.track_command_event(CliEvent.completed)


@meltano_environment.command(cls=PartialInstrumentedCmd, name="list")
@click.pass_context
def list_environments(ctx: click.Context) -> None:
    """List available environments."""
    tracker = ctx.obj["tracker"]
    environment_service: EnvironmentService = ctx.obj[ENVIRONMENT_SERVICE_KEY]
    try:
        for environment in environment_service.list_environments():
            click.echo(environment.name)
    except Exception:
        tracker.track_command_event(CliEvent.failed)
        raise
    tracker.track_command_event(CliEvent.completed)



"""Service for managing task sets (aka jobs)."""

from __future__ import annotations

import typing as t

import structlog

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.task_sets import TaskSets

logger = structlog.getLogger(__name__)


class JobAlreadyExistsError(Exception):
    """A TaskSet (aka job) already exists."""

    def __init__(self, name: str):
        """Initialize a JobAlreadyExistsError.

        Args:
            name: Name of the TaskSet (aka job) that already exists.
        """
        self.name = name


class JobNotFoundError(Exception):
    """A TaskSet (aka job) does not exists."""

    def __init__(self, name: str):
        """Initialize a JobNotFoundError.

        Args:
            name: Name of the TaskSet (aka job) that wasn't found.
        """
        self.name = name
        super().__init__(f"Job '{name}' was not found.")


class TaskSetsService:
    """A service for managing project TaskSets."""

    def __init__(self, project: Project):
        """Initialize a TaskSetsService.

        Args:
            project: The active project.
        """
        self.project = project

    def add(self, task_sets: TaskSets) -> None:
        """Add a TaskSet to the project.

        Args:
            task_sets: The TaskSets to add.

        Raises:
            JobAlreadyExistsError: When a TaskSet with the same name already exists.
        """
        if self.exists(task_sets.name):
            raise JobAlreadyExistsError(task_sets.name)

        with self.project.meltano_update() as meltano:
            logger.debug("adding job", name=task_sets.name, tasks=task_sets.tasks)
            meltano.jobs.append(task_sets)

    def remove(self, name: str) -> TaskSets:
        """Remove a TaskSet from the project.

        Args:
            name: The name of the TaskSet to remove.

        Returns:
            The removed TaskSet.

        Raises:
            JobNotFoundError: If the TaskSet with the given name does not exist.
        """
        for job in self.list():
            if job.name == name:
                with self.project.meltano_update() as meltano:
                    logger.debug("removing job", name=job.name)
                    meltano.jobs.remove(job)
                return job
        raise JobNotFoundError(name)

    def update(self, task_sets: TaskSets) -> None:
        """Update a tasks.

        Args:
            task_sets: The TaskSets to update.

        Raises:
            JobNotFoundError: If the TaskSet with the given name does not exist.
        """
        for idx, job in enumerate(self.list()):
            if job.name == task_sets.name:
                with self.project.meltano_update() as meltano:
                    logger.debug("updating job", name=job.name)
                    meltano.jobs[idx].tasks = task_sets.tasks
                return
        raise JobNotFoundError(task_sets.name)

    def get(self, name: str) -> TaskSets:
        """Get a TaskSet by name.

        Args:
            name: The name of the TaskSet.

        Returns:
            The TaskSet with the given name.

        Raises:
            JobNotFoundError: If the TaskSet with the given name does not exist.
        """
        for job in self.project.meltano.jobs:
            if job.name == name:
                return job
        raise JobNotFoundError(name)

    def list(self) -> list[TaskSets]:
        """List all TaskSets in the project.

        Returns:
            A list of all TaskSets in the project.
        """
        return self.project.meltano.jobs.copy()

    def exists(self, name: str) -> bool:
        """Check if a TaskSet with the given name exists.

        Args:
            name: The name of the TaskSet.

        Returns:
            True if the TaskSet with the given name exists, False otherwise.
        """
        try:
            self.get(name)
        except JobNotFoundError:
            return False
        return True



from __future__ import annotations

import platform
import subprocess
import typing as t

import pytest

from asserts import assert_cli_runner
from meltano.cli import cli
from meltano.core.settings_service import FEATURE_FLAG_PREFIX, FeatureFlags
from meltano.core.utils import EnvironmentVariableNotSetError


class EnvVarResolutionExpectation(t.NamedTuple):
    expected_env_values: dict
    meltanofile_updates: t.ClassVar[dict] = {}
    terminal_env: t.ClassVar[dict] = {}


def _meltanofile_update_dict(
    *,
    top_level_plugin_setting=True,
    top_level_plugin_config=False,
    top_level_env=False,
    top_level_plugin_env=False,
    environment_level_env=False,
    environment_level_plugin_env=False,
    environment_level_plugin_config=False,
    environment_level_plugin_config_indirected=False,
):
    plugin_name = "test-env-var-resolution"
    plugin_namespace = plugin_name.replace("-", "_")
    setting = {
        "name": "from",
        "kind": "string",
    }
    env = {}
    environment = {"name": "dev"}
    utility = {
        "name": plugin_name,
        "namespace": plugin_namespace,
        "settings": [setting],
        "executable": "pwd",
    }
    if top_level_plugin_setting:
        setting["value"] = "top_level_plugin_setting"
    if top_level_plugin_config:
        utility["config"] = {"from": "top_level_plugin_config"}
    if top_level_env:
        env["TEST_ENV_VAR_RESOLUTION_FROM"] = "top_level_env"
    if top_level_plugin_env:
        utility["env"] = {"TEST_ENV_VAR_RESOLUTION_FROM": "top_level_plugin_env"}
    if environment_level_plugin_config:
        environment["config"] = {
            "plugins": {
                "utilities": [
                    {
                        "name": plugin_name,
                        "config": {"from": "environment_level_plugin_config"},
                    },
                ],
            },
        }
    if environment_level_env:
        environment["env"] = {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"}
    if environment_level_plugin_env:
        environment["config"] = {
            "plugins": {
                "utilities": [
                    {
                        "name": plugin_name,
                        "env": {
                            "TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_plugin_env",  # noqa: E501
                        },
                    },
                ],
            },
        }
    if environment_level_plugin_config_indirected:
        environment.update(
            {
                "config": {
                    "plugins": {
                        "utilities": [
                            {
                                "name": plugin_name,
                                "config": {"from": "$INDIRECTED_ENV_VAR"},
                            },
                        ],
                    },
                },
                "env": {
                    "INDIRECTED_ENV_VAR": "environment_level_plugin_config_indirected",
                },
            },
        )
    return {
        "plugins": {"utilities": [utility]},
        "environments": [environment],
        "env": env,
    }


_terminal_env_var = {"TEST_ENV_VAR_RESOLUTION_FROM": "terminal_env"}

_env_var_resolution_expectations = {
    # Check that envs at each level override terminal
    "00 Terminal environment": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "terminal_env"},
        _meltanofile_update_dict(),
        _terminal_env_var,
    ),
    "01 Top-level env (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "top_level_env"},
        _meltanofile_update_dict(top_level_env=True),
        _terminal_env_var,
    ),
    "02 Top-level plugin env (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "top_level_plugin_env"},
        _meltanofile_update_dict(top_level_plugin_env=True),
        _terminal_env_var,
    ),
    "03 Environment-level env (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(environment_level_env=True),
        _terminal_env_var,
    ),
    "04 Environment-level plugin env (with terminal context)": EnvVarResolutionExpectation(  # noqa: E501
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_plugin_env"},
        _meltanofile_update_dict(environment_level_plugin_env=True),
        _terminal_env_var,
    ),
    # Now check the order of precedence between each level
    "06 Top-level env": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "top_level_env"},
        _meltanofile_update_dict(top_level_env=True),
    ),
    "07 Top-level plugin env (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "top_level_plugin_env"},
        _meltanofile_update_dict(top_level_env=True, top_level_plugin_env=True),
    ),
    "08 Environment-level env": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            top_level_env=True,
            top_level_plugin_env=True,
            environment_level_env=True,
        ),
    ),
    "09 Environment-level Plugin env": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_plugin_env"},
        _meltanofile_update_dict(
            top_level_env=True,
            top_level_plugin_env=True,
            environment_level_env=True,
            environment_level_plugin_env=True,
        ),
    ),
    "10 Top-level plugin setting (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "terminal_env"},
        _meltanofile_update_dict(top_level_plugin_setting=True),
        _terminal_env_var,
    ),
    "11 Top-level plugin config (with terminal context)": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
            top_level_plugin_config=True,
        ),
        _terminal_env_var,
    ),
    "12 Environment-level plugin config (with terminal context)": EnvVarResolutionExpectation(  # noqa: E501
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
            top_level_plugin_config=True,
            environment_level_plugin_config=True,
        ),
        _terminal_env_var,
    ),
    "13 Top-level plugin setting": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
        ),
    ),
    "14 Set in top-level plugin config": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
            top_level_plugin_config=True,
        ),
    ),
    "15 Environment-level plugin config": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_env"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
            top_level_plugin_config=True,
            environment_level_plugin_config=True,
        ),
    ),
    "16 Set in indirected environment-level plugin config": EnvVarResolutionExpectation(
        {"TEST_ENV_VAR_RESOLUTION_FROM": "environment_level_plugin_config_indirected"},
        _meltanofile_update_dict(
            environment_level_env=True,
            top_level_plugin_setting=True,
            top_level_plugin_config=True,
            environment_level_plugin_config=True,
            environment_level_plugin_config_indirected=True,
        ),
    ),
}


class TestEnvVarResolution:
    @pytest.mark.parametrize(
        ("expected_env_values", "meltanofile_updates", "terminal_env"),
        [tuple(x) for x in _env_var_resolution_expectations.values()],
        ids=_env_var_resolution_expectations.keys(),
    )
    @pytest.mark.usefixtures("cli_runner")
    def test_env_var_resolution(
        self,
        expected_env_values,
        meltanofile_updates,
        terminal_env,
        project,
        monkeypatch,
    ) -> None:
        if platform.system() == "Windows":
            pytest.xfail(
                "Fails on Windows: https://github.com/meltano/meltano/issues/3444",
            )

        for key, val in terminal_env.items():
            monkeypatch.setenv(key, val)

        with project.meltano_update() as meltanofile:
            meltanofile.update(meltanofile_updates)

        result = subprocess.run(
            (
                "meltano",
                "invoke",
                *(arg for key in expected_env_values for arg in ("--print-var", key)),
                "test-env-var-resolution",
            ),
            text=True,
            stdout=subprocess.PIPE,
            check=True,
        )
        assert result.stdout.strip().split("\n")[:-1] == [
            f"{env_key}={env_val}" for env_key, env_val in expected_env_values.items()
        ]


def test_environment_variable_inheritance(cli_runner, project, monkeypatch) -> None:
    monkeypatch.setenv("STACKED", "1")
    with project.meltano_update() as meltanofile:
        meltanofile.update(
            {
                "env": {"STACKED": "${STACKED}2"},
                "plugins": {
                    "utilities": [
                        {
                            "name": "test-environment-inheritance",
                            "namespace": "test_environment_inheritance",
                            "executable": "pwd",
                            "env": {"STACKED": "${STACKED}4"},
                        },
                    ],
                },
                "environments": [
                    {
                        "name": "dev",
                        "env": {"STACKED": "${STACKED}3"},
                        "config": {
                            "plugins": {
                                "utilities": [
                                    {
                                        "name": "test-environment-inheritance",
                                        "env": {"STACKED": "${STACKED}5"},
                                    },
                                ],
                            },
                        },
                    },
                ],
            },
        )
    result = cli_runner.invoke(
        cli,
        [
            "invoke",
            "--print-var",
            "STACKED",
            "test-environment-inheritance",
        ],
    )
    assert_cli_runner(result)
    assert result.stdout.strip() == "STACKED=12345"


def test_environment_variable_inheritance_meltano_env_only(
    cli_runner,
    project,
    monkeypatch,
) -> None:
    monkeypatch.setenv("STACKED", "1")
    with project.meltano_update() as meltanofile:
        meltanofile.update(
            {
                "plugins": {
                    "utilities": [
                        {
                            "name": "test-environment-inheritance",
                            "namespace": "test_environment_inheritance",
                            "executable": "pwd",
                        },
                    ],
                },
                "environments": [
                    {
                        "name": "dev",
                        "env": {"STACKED": "${STACKED}2"},
                    },
                ],
            },
        )
    result = cli_runner.invoke(
        cli,
        [
            "invoke",
            "--print-var",
            "STACKED",
            "test-environment-inheritance",
        ],
    )
    assert_cli_runner(result)
    assert result.stdout.strip() == "STACKED=12"


def test_strict_env_var_mode_raises_full_replace(cli_runner, project) -> None:
    project.settings.set(
        [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
        value=True,
    )
    with project.meltano_update() as meltanofile:
        meltanofile.update(_meltanofile_update_dict())
        meltanofile.update({"env": {"some_var": "$NONEXISTENT"}})
    result = cli_runner.invoke(
        cli,
        [
            "invoke",
            "--print-var",
            "STACKED",
            "test-env-var-resolution",
        ],
    )
    assert isinstance(result.exception, EnvironmentVariableNotSetError)
    assert (
        result.exception.reason
        == "Environment variable 'NONEXISTENT' referenced but not set"
    )
    assert result.exception.instruction == "Make sure the environment variable is set"


def test_strict_env_var_mode_raises_partial_replace(cli_runner, project) -> None:
    project.settings.set(
        [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
        value=True,
    )
    with project.meltano_update() as meltanofile:
        meltanofile.update(_meltanofile_update_dict())
        meltanofile.update({"env": {"some_var": "some_${NONEXISTENT}_value"}})
    result = cli_runner.invoke(
        cli,
        [
            "invoke",
            "--print-var",
            "STACKED",
            "test-env-var-resolution",
        ],
    )
    exception = result.exception
    assert isinstance(exception, EnvironmentVariableNotSetError)
    assert (
        result.exception.reason
        == "Environment variable 'NONEXISTENT' referenced but not set"
    )
    assert result.exception.instruction == "Make sure the environment variable is set"



from __future__ import annotations

import warnings

import pytest

from meltano.core.environment import Environment
from meltano.core.project_settings_service import (
    ProjectSettingsService,
    SettingValueStore,
)
from meltano.core.settings_service import (
    EXPERIMENTAL,
    FEATURE_FLAG_PREFIX,
    FeatureFlags,
    FeatureNotAllowedException,
)
from meltano.core.utils import EnvironmentVariableNotSetError


@pytest.fixture()
def config_override():
    try:
        ProjectSettingsService.config_override["project_id"] = "from_config_override"

        yield
    finally:
        ProjectSettingsService.config_override.pop("project_id")


@pytest.fixture()
def subject(project):
    return ProjectSettingsService(project)


class TestProjectSettingsService:
    @pytest.fixture()
    def environment(self):
        return Environment("testing", {})

    def test_get_with_source(self, subject, monkeypatch) -> None:
        # A warning is raised because the setting does not exist.
        with pytest.warns(RuntimeWarning):
            assert subject.get_with_source(
                "and_now_for_something_completely_different",
            ) == (None, SettingValueStore.DEFAULT)

        def assert_value_source(value, source) -> None:
            assert subject.get_with_source("project_id") == (value, source)

        subject.set(
            "project_id",
            "from_meltano_yml",
            store=SettingValueStore.MELTANO_YML,
        )

        assert_value_source("from_meltano_yml", SettingValueStore.MELTANO_YML)

        subject.set("project_id", "from_dotenv", store=SettingValueStore.DOTENV)

        assert_value_source("from_dotenv", SettingValueStore.DOTENV)

        with monkeypatch.context() as ctx:
            env_key = subject.setting_env(subject.find_setting("project_id"))
            ctx.setenv(env_key, "from_env")

            assert_value_source("from_env", SettingValueStore.ENV)

    @pytest.mark.usefixtures("config_override")
    def test_get_with_source_config_override(self, subject) -> None:
        assert subject.get_with_source("project_id") == (
            "from_config_override",
            SettingValueStore.CONFIG_OVERRIDE,
        )

    def test_experimental_on(self, subject, monkeypatch) -> None:
        changed = []
        monkeypatch.setenv("MELTANO_EXPERIMENTAL", "true")
        with subject.feature_flag(EXPERIMENTAL):
            changed.append(True)
        assert changed

    def test_experimental_off_by_default(self, subject) -> None:
        changed = []
        with pytest.raises(FeatureNotAllowedException), subject.feature_flag(
            EXPERIMENTAL,
        ):
            changed.append(True)

    @pytest.mark.filterwarnings("ignore:Unknown setting 'ff.allowed':RuntimeWarning")
    def test_feature_flag_allowed(self, subject) -> None:
        changed = []
        subject.set([FEATURE_FLAG_PREFIX, "allowed"], value=True)

        @subject.feature_flag("allowed")
        def should_run() -> None:
            changed.append(True)

        should_run()
        assert changed

    @pytest.mark.filterwarnings("ignore:Unknown setting 'ff.disallowed':RuntimeWarning")
    def test_feature_flag_disallowed(self, subject) -> None:
        changed = []
        subject.set([FEATURE_FLAG_PREFIX, "disallowed"], value=False)

        @subject.feature_flag("disallowed")
        def should_not_run() -> None:
            changed.append(True)

        with pytest.raises(FeatureNotAllowedException):
            should_not_run()

    @pytest.mark.filterwarnings(
        "ignore:Unknown setting 'stacked_env_var':RuntimeWarning",
    )
    def test_strict_env_var_mode_on_raises_error(self, subject) -> None:
        subject.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            value=True,
        )
        subject.set(
            "stacked_env_var",
            "${NONEXISTENT_ENV_VAR}@nonexistent",
        )
        with pytest.raises(EnvironmentVariableNotSetError):
            subject.get("stacked_env_var")

    @pytest.mark.filterwarnings(
        "ignore:Unknown setting 'stacked_env_var':RuntimeWarning",
    )
    def test_strict_env_var_mode_off_no_raise_error(self, subject) -> None:
        subject.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            value=False,
        )
        subject.set(
            "stacked_env_var",
            "${NONEXISTENT_ENV_VAR}@nonexistent_1",
        )
        assert subject.get("stacked_env_var") == "@nonexistent_1"

    def test_warn_if_default_setting_is_used(self, subject, monkeypatch) -> None:
        # Assert that warnings are not raised in the following cases:
        with warnings.catch_warnings():
            warnings.simplefilter("error")

            for ff in FeatureFlags:
                subject.get(f"{FEATURE_FLAG_PREFIX}.{ff}")

            subject.get("project_id")
            subject.get("disable_tracking")

            ProjectSettingsService.config_override.pop(
                "send_anonymous_usage_stats",
                None,
            )
            subject.config_override.pop("send_anonymous_usage_stats", None)
            subject.unset("send_anonymous_usage_stats")

            monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "True")
            assert subject.get("send_anonymous_usage_stats")
            monkeypatch.setenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS", "False")
            assert not subject.get("send_anonymous_usage_stats")
            monkeypatch.delenv("MELTANO_SEND_ANONYMOUS_USAGE_STATS")
            subject.get("send_anonymous_usage_stats")

        # Assert that warnings are raised in the following cases:
        for setting in (
            "projectID",
            "tracking_disabled",
            "MELTANO_SEND_ANONYMOUS_USAGE_STATS",
            "send_anon_usage_stats",
        ):
            with pytest.warns(RuntimeWarning):
                subject.get(setting)

    def test_meltano_settings_with_active_environment(
        self,
        subject,
        monkeypatch,
        environment,
    ) -> None:
        # make sure that meltano setting values are written to the root of `meltano.yml`
        # even if there is an active environment
        monkeypatch.setattr(subject.project, "environment", environment)
        assert subject.project.environment == environment
        subject.set("database_max_retries", 10000)
        value, source = subject.get_with_source("database_max_retries")
        assert source == SettingValueStore.MELTANO_YML
        assert value == 10000

    def test_fully_missing_env_var_setting_is_none(
        self,
        subject: ProjectSettingsService,
    ) -> None:
        subject.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            value=False,
        )
        # https://github.com/meltano/meltano/issues/7189#issuecomment-1396112167
        with pytest.warns(RuntimeWarning, match="Unknown setting 'port'"):
            subject.set("port", "${UNSET_PORT_ENV_VAR}")
        assert subject.get("port") is None

    def test_env_var_settings_expanded_before_cast(
        self,
        subject: ProjectSettingsService,
    ) -> None:
        name = "database_max_retries"  # Using this because it's an int setting
        setting_def = subject.find_setting(name)

        subject.set(name, setting_def.value, setting_def=setting_def)
        assert subject.get(name) == setting_def.value

        SettingValueStore.MELTANO_YML.manager(subject).set(
            name,
            [name],
            "$DB_MAX_RETRIES_TEST",
            setting_def=setting_def,
        )

        subject.set(
            [FEATURE_FLAG_PREFIX, str(FeatureFlags.STRICT_ENV_VAR_MODE)],
            value=False,
        )
        assert subject.get(name) is None

        subject.env_override["DB_MAX_RETRIES_TEST"] = "7"
        assert subject.get(name) == 7



"""ensure user.login_count default value.

Revision ID: 990c0665f3ce
Revises: a3e2b0a4937d
Create Date: 2020-01-15 15:58:25.416129

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import column, table

# revision identifiers, used by Alembic.
revision = "990c0665f3ce"
down_revision = "a3e2b0a4937d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # update the existing Users
    user = table("user", column("login_count", sa.Integer))
    connection = op.get_bind()

    connection.execute(
        user.update().where(user.c.login_count is None).values({"login_count": 0})
    )


def downgrade() -> None:
    pass



from __future__ import annotations

from contextlib import nullcontext

import pytest
from mock import Mock

from meltano.core.db import (
    MeltanoDatabaseCompatibilityError,
    check_database_compatibility,
)


class TestDatabaseCompatibility:
    @pytest.mark.parametrize(
        ("dialect", "version", "expected"),
        (
            pytest.param(
                "sqlite",
                (3, 25, 0),
                pytest.raises(
                    MeltanoDatabaseCompatibilityError,
                    match=(
                        "Detected SQLite 3.25.0, but Meltano requires at least 3.25.1. "
                        "Upgrade your database to be compatible with Meltano or use a "
                        "different database."
                    ),
                ),
                id="sqlite-3.25.0",
            ),
            pytest.param("sqlite", (3, 25, 1), nullcontext(), id="sqlite-3.25.1"),
            pytest.param("sqlite", (3, 26, 0), nullcontext(), id="sqlite-3.26.0"),
            pytest.param("postgresql", (9, 6, 0), nullcontext(), id="postgresql-9.6.0"),
            pytest.param(
                "postgresql",
                (10, 0, 0),
                nullcontext(),
                id="postgresql-10.0.0",
            ),
        ),
    )
    def test_db_compatibility(self, dialect, version, expected) -> None:
        engine_mock = Mock()
        engine_mock.dialect.name = dialect
        engine_mock.dialect.server_version_info = version

        with expected:
            check_database_compatibility(engine_mock)



"""Interactivity utils."""

from __future__ import annotations

from enum import Enum


class InteractionStatus(str, Enum):
    """Interaction status constants."""

    EXIT = "exit"
    SKIP = "skip"
    START = "start"
    RETRY = "retry"



"""Defines SystemModel parent class for other models."""

from __future__ import annotations

import typing as t
from datetime import datetime

from sqlalchemy import MetaData, types
from sqlalchemy.orm import DeclarativeBase

SystemMetadata = MetaData()


class SystemModel(DeclarativeBase):
    """Base class for all database models."""

    metadata = SystemMetadata
    type_annotation_map: t.ClassVar[dict] = {
        str: types.String(),
        bool: types.Boolean(),
        datetime: types.DateTime(),
    }



"""Add 'trigger' to Job.

Revision ID: 367228df6a43
Revises: 87d9638f6ac6
Create Date: 2020-02-19 14:40:30.229756

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    get_dialect_name,
    max_string_length_for_dialect,
)

# revision identifiers, used by Alembic.
revision = "367228df6a43"
down_revision = "87d9638f6ac6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    dialect_name = get_dialect_name()
    max_string_length = max_string_length_for_dialect(dialect_name)

    op.add_column("job", sa.Column("trigger", sa.String(max_string_length)))


def downgrade() -> None:
    op.drop_column("job", "trigger")



"""Click parameter helper decorators."""

from __future__ import annotations

import functools
import typing as t
import uuid

import click

from meltano.cli.utils import AutoInstallBehavior, CliError
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationError
from meltano.core.plugin_install_service import PluginInstallReason, install_plugins
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import async_noop

if t.TYPE_CHECKING:
    from meltano.core.project import Project

_AnyCallable = t.Callable[..., t.Any]
FC = t.TypeVar("FC", bound=_AnyCallable)

InstallPlugins = t.Callable[..., t.Coroutine[t.Any, t.Any, bool]]


def _install_plugins_fn(
    ctx: click.Context,
    _param,
    value: AutoInstallBehavior | None,
) -> InstallPlugins:
    if value is None:
        project: Project | None = ctx.obj["project"]
        auto_install = project and project.settings.get("auto_install")
        behavior = (
            AutoInstallBehavior.install
            if auto_install
            else AutoInstallBehavior.no_install
        )
    else:
        behavior = value

    behavior = value or ctx.obj.get(
        "auto_install_behaviour",
        AutoInstallBehavior.install,
    )
    if behavior == AutoInstallBehavior.no_install:
        return async_noop
    if behavior == AutoInstallBehavior.only_install:
        return _install_plugins_and_exit

    return install_plugins


async def _install_plugins_and_exit(*args, **kwargs) -> bool:  # noqa: ANN002, ANN003
    kwargs.pop("reason", None)
    await install_plugins(*args, **kwargs, reason=PluginInstallReason.INSTALL)
    context = click.get_current_context()
    context.exit(code=0)
    return True  # pragma: no cover


def database_uri_option(func):  # noqa: ANN001, ANN201
    """Database URI Click option decorator.

    Args:
        func: The function to decorate.
    """

    @click.option("--database-uri", help="System database URI.")
    def decorate(*args, database_uri=None, **kwargs):  # noqa: ANN001, ANN002, ANN003, ANN202
        if database_uri:
            ProjectSettingsService.config_override["database_uri"] = database_uri

        return func(*args, **kwargs)

    return functools.update_wrapper(decorate, func)


def get_install_options(
    *,
    include_only_install: bool,
) -> tuple[t.Callable[[FC], FC], ...]:
    """Return install options for CLI commands.

    Args:
        include_only_install: Flag to include the `--only-install` option.

    Returns:
        Tuple of install option decorators.
    """
    install_option = click.option(
        "--install",
        "install_plugins",
        flag_value=AutoInstallBehavior.install.value,
        callback=_install_plugins_fn,
        help="Install the subject plugin(s) automatically.",
    )

    no_install_option = click.option(
        "--no-install",
        "install_plugins",
        flag_value=AutoInstallBehavior.no_install.value,
        callback=_install_plugins_fn,
        help="Do not install the subject plugin(s) automatically.",
    )

    only_install_option = click.option(
        "--only-install",
        "install_plugins",
        flag_value=AutoInstallBehavior.only_install.value,
        callback=_install_plugins_fn,
        help="Only install the subject plugin(s).",
    )

    if include_only_install:
        return install_option, no_install_option, only_install_option

    return install_option, no_install_option


class pass_project:  # noqa: N801
    """Pass current project to decorated CLI command function."""

    __name__ = "project"

    def __init__(self, *, migrate=False) -> None:  # noqa: ANN001
        """Instantiate decorator.

        Args:
            migrate: Flag to perform database migration before passing the project.
        """
        self.migrate = migrate

    def __call__(self, func):  # noqa: ANN001, ANN204
        """Return decorated function.

        Args:
            func: The function to decorate.
        """

        @database_uri_option
        def decorate(*args, **kwargs) -> None:  # noqa: ANN002, ANN003
            ctx = click.get_current_context()

            project = ctx.obj["project"]
            if not project:
                raise CliError(
                    f"`{ctx.command_path}` must be run inside a Meltano project.\n"  # noqa: EM102
                    "Use `meltano init <project_directory>` to create one.",
                )

            # register the system database connection
            engine, _ = project_engine(project, default=True)

            if self.migrate:
                from meltano.core.migration_service import MigrationService

                try:
                    migration_service = MigrationService(engine)
                    migration_service.upgrade(silent=True)
                except MigrationError as err:
                    raise CliError(str(err)) from err

            func(project, *args, **kwargs)

        return functools.update_wrapper(decorate, func)


class UUIDParamType(click.ParamType):
    """A custom click parameter type for UUIDs."""

    name = "uuid"

    def convert(
        self,
        value: str,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> uuid.UUID:
        """Convert an input value to a UUID."""
        try:
            return uuid.UUID(value)
        except ValueError:
            self.fail(f"{value} is not a valid UUID.", param, ctx)



from __future__ import annotations  # noqa: D100

from pathlib import Path

import psutil
import structlog

logger = structlog.stdlib.get_logger(__name__)


class PIDFile:  # noqa: D101
    def __init__(self, path: str | Path):  # noqa: D107
        self.path = Path(path)
        self._pid = None

    @property
    def pid(self):  # noqa: ANN201, D102
        return self._pid or self.load_pid()

    def load_pid(self) -> int:  # noqa: D102
        try:
            with self.path.open("r") as raw:
                self._pid = int(raw.read())
        except ValueError:
            self.path.unlink()
        except FileNotFoundError:
            pass

        logger.debug(f"Loaded PID {self._pid} from {self.path}")  # noqa: G004
        return self._pid

    def write_pid(self, pid: int) -> None:  # noqa: D102
        with self.path.open("w") as raw:
            raw.write(str(pid))
            self._pid = pid

    @property
    def process(self) -> psutil.Process:  # noqa: D102
        if self.pid is None:
            raise UnknownProcessError(self)

        try:
            return psutil.Process(self.pid)
        except psutil.NoSuchProcess as ex:
            raise UnknownProcessError(self) from ex

    def unlink(self):  # noqa: ANN201, D102
        return self.path.unlink()

    def __str__(self) -> str:  # noqa: D105
        return str(self.path)


class UnknownProcessError(Exception):
    """The PIDFile doesn't yield a readable PID."""

    def __init__(self, pid_file: PIDFile):  # noqa: D107
        self.pid_file = pid_file

        super().__init__(f"{pid_file} doesn't refer to a valid Process.")



"""Pytest fixtures that use `pytest-docker` to provide Docker services."""

from __future__ import annotations

from pathlib import Path

import pytest

from .snowplow import SnowplowMicro, snowplow, snowplow_optional, snowplow_session

__all__ = [
    "SnowplowMicro",
    "snowplow",
    "snowplow_optional",
    "snowplow_session",
]


# Originally defined by `pytest-docker`. Overridden to provide a custom location.
@pytest.fixture(scope="session")
def docker_compose_file() -> str:
    """Get the absolute path to the `docker-compose.yml` file used by `pytest-docker`.

    Returns:
        The absolute path to the `docker-compose.yml` file used by `pytest-docker`.
    """
    return str(Path(__file__).parent.resolve() / "docker-compose.yml")



from __future__ import annotations  # noqa: D100

import os
import typing as t
from contextlib import contextmanager

import structlog

from meltano.core.utils import makedirs, slugify

if t.TYPE_CHECKING:
    from pathlib import Path

    from meltano.core.project import Project

logger = structlog.stdlib.get_logger(__name__)
MAX_FILE_SIZE = 2097152  # 2MB max


class MissingJobLogException(Exception):
    """Occurs when `JobLoggingService` can not find a requested log."""


class SizeThresholdJobLogException(Exception):
    """A Job log exceeds the `MAX_FILE_SIZE`."""


class JobLoggingService:  # noqa: D101
    def __init__(self, project: Project):  # noqa: D107
        self.project = project

    @makedirs
    def logs_dir(self, state_id, *joinpaths, make_dirs: bool = True):  # noqa: ANN001, ANN002, ANN201
        """Return the logs directory for a given state_id.

        Args:
            state_id: The state ID for the logs.
            joinpaths: Additional paths to join to the logs directory.
            make_dirs: Whether to create the directory if it does not exist.

        Returns:
            The logs directory for the given state ID.
        """
        return self.project.job_logs_dir(state_id, *joinpaths, make_dirs=make_dirs)

    def generate_log_name(
        self,
        state_id: str,
        run_id: str,
        file_name: str = "elt.log",
    ) -> Path:
        """Generate an internal etl log path and name.

        Args:
            state_id: The state ID for the log.
            run_id: The run ID for the log.
            file_name: The name of the log file.

        Returns:
            The full path to the log file.
        """
        return self.logs_dir(state_id, str(run_id), file_name)

    @contextmanager
    def create_log(self, state_id, run_id, file_name="elt.log"):  # noqa: ANN001, ANN201
        """Open a new log file for logging and yield it.

        Log will be created inside the logs_dir, which is
        `.meltano/logs/elt/:state_id/:run_id`
        """
        log_file_name = self.generate_log_name(state_id, run_id, file_name)

        try:
            with log_file_name.open("w") as log_file:
                yield log_file
        except OSError:
            # Don't stop the Job running if you can not open the log file
            # for writing: just return /dev/null
            logger.error(
                f"Could open log file {log_file_name!r} for writing. "  # noqa: G004
                "Using `/dev/null`",
            )
            with open(os.devnull, "w") as log_file:  # noqa: PTH123
                yield log_file

    def get_latest_log(self, state_id) -> str:  # noqa: ANN001
        """Get the latest log.

        Args:
            state_id: The state ID for the log.

        Returns:
            The contents of the most recent log for any ELT job that ran with
            the provided `state_id`.
        """
        try:
            latest_log = next(iter(self.get_all_logs(state_id)))

            if latest_log.stat().st_size > MAX_FILE_SIZE:
                raise SizeThresholdJobLogException(
                    f"The log file size exceeds '{MAX_FILE_SIZE}'",  # noqa: EM102
                )

            with latest_log.open() as f:
                return f.read()
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with ID '{state_id}'",  # noqa: EM102
            ) from None
        except FileNotFoundError as ex:
            raise MissingJobLogException(
                f"Cannot log for job with ID '{state_id}': '{latest_log}' is missing.",  # noqa: EM102
            ) from ex

    def get_downloadable_log(self, state_id):  # noqa: ANN001, ANN201
        """Get the `*.log` file of the most recent log for any ELT job that ran with the provided `state_id`."""  # noqa: E501
        try:
            latest_log = next(iter(self.get_all_logs(state_id)))
            return str(latest_log.resolve())
        except StopIteration:
            raise MissingJobLogException(
                f"Could not find any log for job with ID '{state_id}'",  # noqa: EM102
            ) from None
        except FileNotFoundError as ex:
            raise MissingJobLogException(
                f"Cannot log for job with ID '{state_id}': '{latest_log}' is missing.",  # noqa: EM102
            ) from ex

    def get_all_logs(self, state_id):  # noqa: ANN001, ANN201
        """Get all the log files for any ELT job that ran with the provided `state_id`.

        The result is ordered so that the most recent is first on the list.
        """
        return sorted(
            [
                log_file
                for logs_dir in self.logs_dirs(state_id)
                for log_file in logs_dir.glob("**/*.log")
            ],
            key=lambda path: path.stat().st_ctime_ns,
            reverse=True,
        )

    def delete_all_logs(self, state_id) -> None:  # noqa: ANN001
        """Delete all the logs for any ELT job that ran with the provided `state_id`.

        Args:
            state_id: The state ID for which all log files should be deleted.
        """
        for log_path in self.get_all_logs(state_id):
            log_path.unlink()

    def legacy_logs_dir(self, state_id, *joinpaths):  # noqa: ANN001, ANN002, ANN201, D102
        job_dir = self.project.run_dir("elt").joinpath(slugify(state_id), *joinpaths)
        return job_dir if job_dir.exists() else None

    def logs_dirs(self, state_id, *joinpaths):  # noqa: ANN001, ANN002, ANN201, D102
        logs_dir = self.logs_dir(state_id, *joinpaths)
        legacy_logs_dir = self.legacy_logs_dir(state_id, *joinpaths)

        dirs = [logs_dir]
        if legacy_logs_dir:
            dirs.append(legacy_logs_dir)

        return dirs



from __future__ import annotations

import traceback
import typing as t

if t.TYPE_CHECKING:
    from click.testing import Result


def assert_cli_runner(results: Result, message=None) -> None:
    assertion_message = str(message or results.output)

    if results.exception:
        assertion_message += "".join(
            ("\n", *traceback.format_exception(*results.exc_info)),
        )

    assert results.exit_code == 0, assertion_message



"""add last activity at column.

Revision ID: e4fbabc3fed6
Revises: 367228df6a43
Create Date: 2020-03-24 15:53:54.142685

"""

from __future__ import annotations

import sqlalchemy as sa
from alembic import op

from meltano.migrations.utils.dialect_typing import (
    datetime_for_dialect,
    get_dialect_name,
)

# revision identifiers, used by Alembic.
revision = "e4fbabc3fed6"
down_revision = "367228df6a43"
branch_labels = None
depends_on = None


def upgrade() -> None:
    dialect_name = get_dialect_name()
    datetime_type = datetime_for_dialect(dialect_name)

    op.add_column("user", sa.Column("last_activity_at", datetime_type, nullable=True))


def downgrade() -> None:
    op.drop_column("user", "last_activity_at")



"""Job management CLI."""

from __future__ import annotations

import json
import typing as t

import click
import structlog

from meltano.cli.params import pass_project
from meltano.cli.utils import (
    CliEnvironmentBehavior,
    CliError,
    InstrumentedGroup,
    PartialInstrumentedCmd,
)
from meltano.core.block.parser import BlockParser, validate_block_sets
from meltano.core.task_sets import InvalidTasksError, TaskSets, tasks_from_yaml_str
from meltano.core.task_sets_service import (
    JobAlreadyExistsError,
    JobNotFoundError,
    TaskSetsService,
)
from meltano.core.tracking.contexts import CliEvent, PluginsTrackingContext

if t.TYPE_CHECKING:
    from meltano.core.project import Project
    from meltano.core.tracking import Tracker

logger = structlog.getLogger(__name__)


def _list_single_job(  # noqa: D417
    task_sets_service: TaskSetsService,
    list_format: str,
    job_name: str,
    ctx: click.Context,
) -> None:
    """List a single job.

    Args:
        project: The project to use.
        task_sets_service: The task sets service to use.
        list_format: The format to use.
        job_name: The job name to list.
    """
    tracker: Tracker = ctx.obj["tracker"]
    try:
        task_set = task_sets_service.get(job_name)
    except JobNotFoundError:
        tracker.track_command_event(CliEvent.failed)
        click.secho(f"Job '{job_name}' does not exist.", fg="yellow")
        return

    if list_format == "text":
        click.echo(f"{task_set.name}: {task_set.tasks}")
    elif list_format == "json":
        click.echo(
            json.dumps({"job_name": task_set.name, "tasks": task_set.tasks}, indent=2),
        )
    tracker.track_command_event(CliEvent.completed)


def _list_all_jobs(  # noqa: D417
    task_sets_service: TaskSetsService,
    list_format: str,
    ctx: click.Context,
) -> None:
    """List all jobs.

    Args:
        project: The project to use.
        task_sets_service: The task sets service to use.
        list_format: The format to use.
    """
    if list_format == "json":
        click.echo(
            json.dumps(
                {
                    "jobs": [
                        {"job_name": tset.name, "tasks": tset.tasks}
                        for tset in task_sets_service.list()
                    ],
                },
                indent=2,
            ),
        )
    elif list_format == "text":
        for task_set in task_sets_service.list():
            click.echo(f"{task_set.name}: {task_set.tasks}")
    tracker: Tracker = ctx.obj["tracker"]
    tracker.track_command_event(CliEvent.completed)


@click.group(
    cls=InstrumentedGroup,
    short_help="Manage jobs.",
    environment_behavior=CliEnvironmentBehavior.environment_optional_ignore_default,
)
@click.pass_context
@pass_project(migrate=True)
def job(project, ctx) -> None:  # noqa: ANN001
    """Manage jobs.

    Example usage:

    \b
    \t# This help
    \tmeltano job --help
    \t# List all jobs in JSON format
    \tmeltano job list --format json
    \t# List a named job
    \tmeltano job list <job_name>
    \b
    \t# Create a new job with a single task representing a single run command.
    \tmeltano job add NAME --tasks 'tap mapper target command:arg1'
    \b
    \t# Create a new job with multiple tasks each representing a run command.
    \t# The list of tasks must be yaml formatted and consist of a list of strings, list of string lists, or mix of both.
    \tmeltano job add NAME --tasks '["tap mapper target", "tap2 target2", ...]'
    \tmeltano job add NAME --tasks '[["tap target dbt:run", "tap2 target2", ...], ...]'
    \b
    \t# Remove a named job
    \tmeltano job remove <job_name>

    \b
    Read more at https://docs.meltano.com/reference/command-line-interface#jobs
    """  # noqa: D301, E501
    ctx.obj["project"] = project
    ctx.obj["task_sets_service"] = TaskSetsService(project)


@job.command(cls=PartialInstrumentedCmd, name="list", short_help="List job(s).")
@click.option(
    "--format",
    "list_format",
    type=click.Choice(["json", "text"]),
    default="text",
)
@click.argument("job_name", required=False, default=None)
@click.pass_context
def list_jobs(ctx, list_format: str, job_name: str) -> None:  # noqa: ANN001
    """List available jobs."""
    task_sets_service: TaskSetsService = ctx.obj["task_sets_service"]

    if job_name:
        _list_single_job(task_sets_service, list_format, job_name, ctx)
    else:
        _list_all_jobs(task_sets_service, list_format, ctx)


@job.command(
    cls=PartialInstrumentedCmd,
    name="add",
    short_help="Add a new job with tasks.",
)
@click.argument(
    "job_name",
    required=True,
    default=None,
)
@click.option(
    "--tasks",
    "raw_tasks",
    required=True,
    default=None,
    help="Tasks that will be run as part of this job.",
)
@click.pass_context
def add(ctx, job_name: str, raw_tasks: str) -> None:  # noqa: ANN001
    """Add tasks to a new job.

    Example usage:

    \b
    \t# Create a new job with a single task representing a single run command.
    \tmeltano job add NAME --tasks 'tap mapper target command:arg1'
    \b
    \t# Create a new job with multiple tasks each representing a run command.
    \t# The list of tasks must be yaml formatted and consist of a list of strings, list of string lists, or mix of both.
    \tmeltano job add NAME --tasks '["tap mapper target", "tap2 target2", ...]'
    \tmeltano job add NAME --tasks '[["tap target dbt:run", "tap2 target2", ...], ...]'
    """  # noqa: D301, E501
    task_sets_service: TaskSetsService = ctx.obj["task_sets_service"]
    tracker: Tracker = ctx.obj["tracker"]
    project: Project = ctx.obj["project"]

    try:
        task_sets = tasks_from_yaml_str(job_name, raw_tasks)
    except InvalidTasksError as yerr:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(yerr) from yerr

    try:
        _validate_tasks(project, task_sets, ctx)
    except InvalidTasksError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(err) from err

    try:
        task_sets_service.add(task_sets)
    except JobAlreadyExistsError as serr:
        tracker.track_command_event(CliEvent.failed)
        raise CliError(f"Job '{task_sets.name}' already exists.") from serr  # noqa: EM102

    click.echo(f"Added job {task_sets.name}: {task_sets.tasks}")

    tracker.track_command_event(CliEvent.completed)


@job.command(
    cls=PartialInstrumentedCmd,
    name="set",
    short_help="Update an existing jobs tasks",
)
@click.argument(
    "job_name",
    required=True,
    default=None,
)
@click.option(
    "--tasks",
    "raw_tasks",
    required=True,
    default=None,
    help="Tasks that will be run as part of this job.",
)
@click.pass_context
def set_cmd(ctx, job_name: str, raw_tasks: str) -> None:  # noqa: ANN001
    """Update the tasks associated with an existing job.

    Example usage:

    \b
    \t# Update a job with a single task representing a single run command.
    \tmeltano job set NAME --tasks 'tap mapper target command:arg1'
    \b
    \t# Update a job with multiple tasks each representing a run command.
    \t# The list of tasks must be yaml formatted and consist of a list of strings, list string lists, or mix of both.
    \tmeltano job set NAME --tasks '["tap mapper target", "tap2 target2", ...]'
    \tmeltano job set NAME --tasks '[["tap target dbt:run", "tap2 target2", ...], ...]'
    """  # noqa: D301, E501
    tracker: Tracker = ctx.obj["tracker"]
    project: Project = ctx.obj["project"]
    task_sets_service: TaskSetsService = ctx.obj["task_sets_service"]

    task_sets = tasks_from_yaml_str(job_name, raw_tasks)

    try:
        _validate_tasks(project, task_sets, ctx)
    except InvalidTasksError as err:
        tracker.track_command_event(CliEvent.aborted)
        raise CliError(err) from err

    try:
        task_sets_service.update(task_sets)
    except JobNotFoundError:
        tracker.track_command_event(CliEvent.failed)
        click.secho(f"Job '{job_name}' does not exist.", fg="yellow")
        return

    click.echo(f"Updated job {task_sets.name}: {task_sets.tasks}")
    tracker.track_command_event(CliEvent.completed)


@job.command(cls=PartialInstrumentedCmd, name="remove", short_help="Remove a job.")
@click.argument("job_name", required=True)
@click.pass_context
def remove(ctx, job_name: str) -> None:  # noqa: ANN001
    """Remove a job.

    Usage:
        meltano job remove <job_name>
    """
    tracker: Tracker = ctx.obj["tracker"]
    task_sets_service: TaskSetsService = ctx.obj["task_sets_service"]
    task_sets = task_sets_service.remove(job_name)
    click.echo(f"Removed job '{task_sets.name}'.")
    tracker.track_command_event(CliEvent.completed)


def _validate_tasks(project: Project, task_set: TaskSets, ctx: click.Context) -> bool:  # noqa: D417
    """Validate a job's tasks.

    Validates the tasks by attempting to parse them into valid `Blocks`, and by
    using the `Block` validation logic.

    Args:
        project: Project to use.
        task_set: TaskSets to validate.

    Returns:
        True if the job's tasks are valid.

    Raises:
        InvalidTasksError: If the job's tasks are invalid.
    """
    logger.debug("validating job tasks", job=task_set.name, tasks=task_set.tasks)
    tracker: Tracker = ctx.obj["tracker"]
    for task in task_set.flat_args_per_set:
        blocks = task
        logger.debug(
            "validating tasks",
            job=task_set.name,
            task=task,
            blocks=blocks,
        )
        try:
            block_parser = BlockParser(logger, project, blocks)  # type: ignore[arg-type]
            parsed_blocks = list(block_parser.find_blocks(0))
            tracker.add_contexts(PluginsTrackingContext.from_blocks(parsed_blocks))
        except Exception as err:
            tracker.track_command_event(CliEvent.aborted)
            raise InvalidTasksError(task_set.name, str(err)) from err
        if not validate_block_sets(logger, parsed_blocks):
            tracker.track_command_event(CliEvent.aborted)
            raise InvalidTasksError(
                task_set.name,
                "BlockSet validation failed.",
            )
    return True



from __future__ import annotations

import pytest

from meltano.core.behavior.canonical import Canonical
from meltano.core.plugin.command import Command, UndefinedEnvVarError


class TestCommand:
    @pytest.fixture()
    def commands(self):
        return {
            "foo": {"args": "foo", "description": "foo desc", "executable": "foo"},
            "bar": {"args": "bar"},
            "baz": "baz",
            "test": {"args": "--test", "description": "Run tests"},
        }

    def test_serialize(self, commands) -> None:
        assert Command.parse(commands["foo"]).args == "foo"
        assert Command.parse(commands["bar"]).args == "bar"
        assert Command.parse(commands["baz"]).args == "baz"
        assert Command.parse(commands["test"]).args == "--test"

        serialized = Command.parse_all(commands)
        assert serialized["foo"].args == "foo"
        assert serialized["foo"].description == "foo desc"
        assert serialized["foo"].executable == "foo"
        assert serialized["bar"].args == "bar"
        assert serialized["bar"].description is None
        assert serialized["bar"].executable is None
        assert serialized["baz"].args == "baz"
        assert serialized["baz"].description is None
        assert serialized["baz"].executable is None
        assert serialized["test"].args == "--test"
        assert serialized["test"].description == "Run tests"
        assert serialized["test"].executable is None

    def test_deserialize(self, commands) -> None:
        serialized = Command.parse_all(commands)

        assert serialized["foo"].canonical() == commands["foo"]
        assert serialized["bar"].canonical() == "bar"
        assert serialized["baz"].canonical() == "baz"
        assert serialized["test"].canonical() == commands["test"]

        assert Canonical.as_canonical(serialized) == {**commands, "bar": "bar"}

    def test_expanded_args(self) -> None:
        expanded_args = Command.parse("some args --flag $ENV_VAR_ARG").expanded_args(
            name="cmd",
            env={
                "ENV_VAR_ARG": "env-var-arg",
            },
        )

        assert expanded_args == ["some", "args", "--flag", "env-var-arg"]

    def test_undefined_env_var(self) -> None:
        with pytest.raises(UndefinedEnvVarError):
            Command.parse("some args --flag $ENV_VAR_ARG").expanded_args(
                name="cmd",
                env={},
            )



"""Doc CLI command."""

from __future__ import annotations

import click

from meltano.cli.utils import InstrumentedCmd


@click.command(
    cls=InstrumentedCmd,
    short_help="Open the Meltano docs in your browser.",
)
def docs() -> None:
    """Open the Meltano docs in the system browser."""
    click.launch("https://docs.meltano.com/")  # pragma: no cover
    click.secho("Opening the docs...", fg="green")



from __future__ import annotations  # noqa: D100

import asyncio
import sys
import typing as t

from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType

from . import Runner, RunnerError

if t.TYPE_CHECKING:
    from meltano.core.elt_context import ELTContext
    from meltano.core.plugin_invoker import PluginInvoker


class DbtRunner(Runner):  # noqa: D101
    def __init__(self, elt_context: ELTContext):  # noqa: D107
        self.context = elt_context

    @property
    def project(self):  # noqa: ANN201, D102
        return self.context.project

    @property
    def plugin_context(self):  # noqa: ANN201, D102
        return self.context.transformer

    async def invoke(self, dbt: PluginInvoker, *args, log=None, **kwargs) -> None:  # noqa: ANN001, ANN002, ANN003
        """Call the dbt executable with the given arguments in a new process."""
        log = log or sys.stderr

        try:
            handle = await dbt.invoke_async(
                *args,
                **kwargs,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
        except Exception as err:
            raise RunnerError(f"Cannot start dbt: {err}") from err  # noqa: EM102

        await asyncio.wait(
            [
                asyncio.create_task(coro)
                for coro in (
                    capture_subprocess_output(handle.stdout, log),
                    capture_subprocess_output(handle.stderr, log),
                    handle.wait(),
                )
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

        if exitcode := handle.returncode:
            command = kwargs["command"] or args[0]
            raise RunnerError(
                f"`dbt {command}` failed",  # noqa: EM102
                {PluginType.TRANSFORMERS: exitcode},
            )

    async def run(self, log=None) -> None:  # noqa: ANN001, D102
        dbt = self.context.transformer_invoker()

        async with dbt.prepared(self.context.session):
            await self.invoke(dbt, log=log, command="clean")
            await self.invoke(dbt, log=log, command="deps")

            cmd = "compile" if self.context.dry_run else "run"
            await self.invoke(dbt, log=log, command=cmd)



"""Storage Managers for Meltano Configuration."""

from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from contextlib import contextmanager, suppress
from copy import deepcopy
from enum import Enum
from functools import reduce
from operator import eq

import dotenv
import sqlalchemy
import structlog

from meltano.core.environment import NoActiveEnvironment
from meltano.core.error import MeltanoError, ProjectReadonly
from meltano.core.setting import Setting
from meltano.core.setting_definition import SettingDefinition, SettingMissingError
from meltano.core.utils import flatten, pop_at_path, set_at_path

if t.TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from meltano.core.settings_service import SettingsService


logger = structlog.stdlib.get_logger(__name__)


class ConflictingSettingValueException(Exception):
    """A setting has multiple conflicting values via aliases."""

    def __init__(self, setting_names) -> None:  # noqa: ANN001
        """Instantiate the error.

        Args:
            setting_names: the name/aliases where conflicting values are set

        """
        self.setting_names = setting_names
        super().__init__(setting_names)

    def __str__(self) -> str:
        """Represent the error as a string.

        Returns:
            string representation of the error
        """
        return f"Conflicting values for setting found in: {self.setting_names}"


class MultipleEnvVarsSetException(Exception):
    """A setting value is set via multiple environment variable names."""

    def __init__(self, names) -> None:  # noqa: ANN001
        """Instantiate the error.

        Args:
            names: the name/aliases where conflicting values are set
        """
        self.names = names
        super().__init__(names)

    def __str__(self) -> str:
        """Represent the error as a string.

        Returns:
            string representation of the error
        """
        return (
            f"Error: Setting value set via multiple environment variables: {self.names}"
        )


class StoreNotSupportedError(MeltanoError):
    """Error raised when write actions are performed on a Store that is not writable."""

    def __init__(
        self,
        reason: str | Exception = "Store is not supported",
        **kwds: t.Any,
    ) -> None:
        """Instantiate the error."""
        return super().__init__(reason, **kwds)


def cast_setting_value(
    value: t.Any,  # noqa: ANN401
    metadata: dict[str, t.Any],
    setting_def: SettingDefinition | None,
) -> tuple[t.Any, dict[str, t.Any]]:
    """Cast a setting value according to its setting definition.

    Args:
        value: The setting value to be cast.
        metadata: The metadata of the setting value.
        setting_def: The setting definition.

    Returns:
        A tuple with the setting value, and its metadata. If the
        setting definition was `None`, or the value was not changed by
        the cast, the pair returned is the same `value` and `metadata`
        objects provided. Otherwise, the cast value is returned along
        with a new dictionary which is a shallow copy of the provided
        metadata with the "uncast_value" key-value pair added.
    """
    if setting_def is not None:
        cast_value = setting_def.cast_value(value)
        if cast_value != value:
            return cast_value, {**metadata, "uncast_value": value}
    return value, metadata


class SettingValueStore(str, Enum):
    """Setting Value Store.

    Note: The declaration order of stores determines store precedence when using
        the Auto store manager. This is because the `.readables()` and
        `.writables()` methods return ordered lists that the Auto store manager
        iterates over when retrieving setting values.
    """

    CONFIG_OVERRIDE = "config_override"
    ENV = "env"
    DOTENV = "dotenv"
    MELTANO_ENV = "meltano_environment"
    MELTANO_YML = "meltano_yml"
    DB = "db"
    INHERITED = "inherited"
    DEFAULT = "default"
    AUTO = "auto"

    @classmethod
    def readables(cls) -> list[SettingValueStore]:
        """Return list of readable SettingValueStore instances.

        Returns:
            An list of readable stores in order of precedence.
        """
        return list(cls)

    @classmethod
    def writables(cls) -> list[SettingValueStore]:
        """Return list of writable SettingValueStore instances.

        Returns:
            An list of writable stores in order of precedence.
        """
        return [store for store in cls if store.writable]

    @property
    def manager(self) -> type[SettingsStoreManager]:
        """Return store manager for this store.

        Returns:
            SettingsStoreManager for this store.
        """
        # ordering here is not significant, other than being consistent with
        # the order of precedence.
        managers = {
            self.CONFIG_OVERRIDE: ConfigOverrideStoreManager,
            self.ENV: EnvStoreManager,
            self.DOTENV: DotEnvStoreManager,
            self.MELTANO_ENV: MeltanoEnvStoreManager,
            self.MELTANO_YML: MeltanoYmlStoreManager,
            self.DB: DbStoreManager,
            self.INHERITED: InheritedStoreManager,
            self.DEFAULT: DefaultStoreManager,
            self.AUTO: AutoStoreManager,
        }
        return managers[self]

    @property
    def label(self) -> str:
        """Return printable label.

        Returns:
            Printable label.
        """
        return self.manager.label

    @property
    def writable(self) -> bool:
        """Return if this store is writeable.

        Returns:
            True if store is writable.
        """
        return self.manager.writable

    def overrides(self, source: SettingValueStore) -> bool:
        """Check if given source overrides this instance.

        Args:
            source: SettingValueStore to compare against.

        Returns:
            True if given source takes precedence over this store.
        """
        stores_list = list(self.__class__)
        return stores_list.index(self) < stores_list.index(source)

    def can_overwrite(self, source: SettingValueStore) -> bool:
        """Check if source can overwrite.

        Args:
            source: SettingValueStore instance to check.

        Returns:
            True if instance is overwritable.
        """
        return self.writable and (source is self or self.overrides(source))


class SettingsStoreManager(ABC):
    """Base settings store manager."""

    readable = True
    writable = False

    def __init__(
        self,
        settings_service: SettingsService,
        **kwargs,  # noqa: ANN003, ARG002
    ):
        """Initialise settings store manager.

        Args:
            settings_service: SettingsService instance.
            kwargs: Keyword arguments.
        """
        self.settings_service = settings_service
        self.project = self.settings_service.project

    @abstractmethod
    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,
    ) -> tuple[str, dict]:
        """Abstract get method.

        Args:
            name: Setting name.
            setting_def: SettingDefinition instance.
            cast_value: Whether to cast the value according to `setting_def`.
        """

    def set(
        self,
        name: str,
        path: list[str],
        value: t.Any,  # noqa: ANN401
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Unimplemented set method.

        Args:
            name: Setting name.
            path: Setting path.
            value: New value to set.
            setting_def: SettingDefinition instance.

        Raises:
            NotImplementedError: always.
        """
        raise NotImplementedError

    def unset(
        self,
        name: str,
        path: list[str],
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Unimplemented unset method.

        Args:
            name: Setting name.
            path: Setting path.
            setting_def: SettingDefinition instance.

        Raises:
            NotImplementedError: always.
        """
        raise NotImplementedError

    def reset(self) -> dict:
        """Unimplemented reset method.

        Raises:
            NotImplementedError: always.
        """
        raise NotImplementedError

    def ensure_supported(self, method: str = "get") -> None:
        """Ensure passed method is supported.

        Args:
            method: Method to check if supported.

        Raises:
            StoreNotSupportedError: if method is not "get" a store is not writable.
        """
        if method != "get" and not self.writable:
            reason = "Store is not writable"
            instruction = f"Use one of {set(SettingValueStore.writables())}."
            raise StoreNotSupportedError(
                reason,
                instruction=instruction,
            )

    def log(self, message: str) -> None:
        """Log method.

        Args:
            message: message to log.
        """
        self.settings_service.log(message)


class ConfigOverrideStoreManager(SettingsStoreManager):
    """Config override store manager."""

    label = "a command line flag"

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,  # noqa: ARG002
        *,
        cast_value: bool = False,  # noqa: ARG002
    ) -> tuple[str, dict]:
        """Get value by name from the .env file.

        Args:
            name: Setting name.
            setting_def: Unused. Included to match parent class method signature.
            cast_value: Whether to cast the value according to `setting_def`.

        Returns:
            A tuple the got value and an empty dictionary.
        """
        try:
            value = self.settings_service.config_override[name]
            self.log(f"Read key '{name}' from config override: {value!r}")
            return value, {}
        except KeyError:
            return None, {}


class BaseEnvStoreManager(SettingsStoreManager):
    """Base terminal environment store manager."""

    @property
    @abstractmethod
    def env(self):  # noqa: ANN201
        """Abstract environment values property."""

    def get(
        self,
        name: str,  # noqa: ARG002
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,
    ) -> tuple[str, dict]:
        """Get value by name from the .env file.

        Args:
            name: Unused. Included to match parent class method signature.
            setting_def: SettingDefinition instance.
            cast_value: Whether to cast the value according to `setting_def`.

        Raises:
            StoreNotSupportedError: if setting_def not passed.
            ConflictingSettingValueException: if multiple conflicting values for the
                same setting are provided.
            MultipleEnvVarsSetException: if multiple environment variables are set for
                the same setting.

        Returns:
            A tuple the got value and a dictionary containing metadata.
        """
        if not setting_def:
            reason = "Can not retrieve unknown setting from environment variables"
            raise StoreNotSupportedError(reason)

        vals_with_metadata = []
        for env_var in self.setting_env_vars(setting_def):
            with suppress(KeyError):
                value = env_var.get(self.env)
                vals_with_metadata.append((value, {"env_var": env_var.key}))
        if len(vals_with_metadata) > 1:
            if reduce(eq, (val for val, _ in vals_with_metadata)):
                raise MultipleEnvVarsSetException(
                    [metadata["env_var"] for _, metadata in vals_with_metadata],
                )
            raise ConflictingSettingValueException(
                [metadata["env_var"] for _, metadata in vals_with_metadata],
            )

        value, metadata = vals_with_metadata[0] if vals_with_metadata else (None, {})
        return (
            cast_setting_value(value, metadata, setting_def)
            if cast_value
            else (value, metadata)
        )

    def setting_env_vars(self, *args, **kwargs) -> dict:  # noqa: ANN002, ANN003
        """Return setting environment variables.

        Args:
            args: Positional arguments to pass to `settings_service.setting_env_vars`.
            kwargs: Keyword arguments to pass to `settings_service.setting_env_vars`.

        Returns:
            A dictionary of setting environment variables.
        """
        return self.settings_service.setting_env_vars(*args, **kwargs)


class EnvStoreManager(BaseEnvStoreManager):
    """Terminal environment store manager."""

    label = "the environment"

    @property
    def env(self):  # noqa: ANN201
        """Return values from the calling terminals environment.

        Returns:
            Values found in the calling terminals environment.
        """
        return self.settings_service.env

    def get(self, *args, **kwargs):  # noqa: ANN002, ANN003, ANN201
        """Get value by name from the .env file.

        Args:
            args: Positional arguments to pass to parent method.
            kwargs: Keyword arguments to pass to parent method.

        Returns:
            A tuple the got value and a dictionary containing metadata.
        """
        value, metadata = super().get(*args, **kwargs)

        if value is not None:
            env_key = metadata["env_var"]
            self.log(f"Read key '{env_key}' from the environment: {value!r}")

        return value, metadata


class DotEnvStoreManager(BaseEnvStoreManager):
    """.env file store manager."""

    label = "`.env`"
    writable = True

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialise a .env file store manager instance.

        Args:
            args: Positional arguments to pass to parent class.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(*args, **kwargs)
        self._env = None

    def ensure_supported(self, method: str = "get") -> None:
        """Ensure named method is supported.

        Checks that the project is not read-only and an environment is active.

        Args:
            method: Setting method (get, set, etc.)

        Raises:
            StoreNotSupportedError: If the project is read-only or named method
                is not "get".
        """
        if method != "get" and self.project.readonly:
            raise StoreNotSupportedError(ProjectReadonly())

    @property
    def env(self) -> dict:
        """Return values from the .env file.

        Returns:
            A dictionary of values found in the .env file.
        """
        if self._env is None:
            self._env = self.project.dotenv_env
        return self._env

    def get(self, *args, **kwargs) -> tuple[str, dict]:  # noqa: ANN002, ANN003
        """Get value by name from the .env file.

        Args:
            args: Positional arguments to pass to parent method.
            kwargs: Keyword arguments to pass to parent method.

        Returns:
            A tuple the got value and a dictionary containing metadata.
        """
        value, metadata = super().get(*args, **kwargs)

        if value is not None:
            env_key = metadata["env_var"]
            self.log(f"Read key '{env_key}' from `.env`: {value!r}")

        return value, metadata

    def set(self, name: str, path: list[str], value, setting_def=None):  # noqa: ANN001, ANN201, ARG002
        """Set value by name in the .env file.

        Args:
            name: Unused. Included to match parent class method signature.
            path: Unused. Included to match parent class method signature.
            value: New value to set.
            setting_def: SettingDefinition.

        Raises:
            StoreNotSupportedError: if setting_def not provided.

        Returns:
            An empty dictionary.
        """
        if not setting_def:
            reason = f"Unknown setting '{name}' can not be set in `.env`"
            instruction = (
                "Define the setting in 'meltano.yml' or use a different setting store"
            )
            raise StoreNotSupportedError(reason, instruction=instruction)

        primary_var, *other_vars = self.setting_env_vars(setting_def)
        primary_key = primary_var.key
        other_keys = [var.key for var in other_vars]

        with self.update_dotenv() as dotenv_file:
            if dotenv_file.exists():
                for key in other_keys:
                    dotenv.unset_key(dotenv_file, key)
                    self.log(f"Unset key '{key}' in `.env`")
            else:
                dotenv_file.touch()

            dotenv.set_key(dotenv_file, primary_key, setting_def.stringify_value(value))

        self.log(f"Set key '{primary_key}' in `.env`: {value!r}")
        return {"env_var": primary_key}

    def unset(
        self,
        name: str,
        path: list[str],  # noqa: ARG002
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Unset value by SettingDefinition in the .env file.

        Args:
            name: Unused. Included to match parent class method signature.
            path: Unused. Included to match parent class method signature.
            setting_def: SettingDefinition instance.

        Returns:
            An empty dictionary.

        Raises:
            StoreNotSupportedError: if setting_def not passed.
        """
        if not setting_def:
            reason = f"Unknown setting {name} can not be unset in `.env`"
            instruction = (
                "Define the setting in 'meltano.yml' or use a different setting store"
            )
            raise StoreNotSupportedError(reason, instruction=instruction)

        env_vars = self.setting_env_vars(setting_def)
        env_keys = [var.key for var in env_vars]

        with self.update_dotenv() as dotenv_file:
            if not dotenv_file.exists():
                return {}

            for key in env_keys:
                dotenv.unset_key(dotenv_file, key)
                self.log(f"Unset key '{key}' in `.env`")

        return {}

    def reset(self) -> dict:
        """Reset all Setting values in this store.

        Returns:
            An empty dictionary.
        """
        with self.update_dotenv() as dotenv_file:
            if dotenv_file.exists():
                dotenv_file.unlink()

        return {}

    @contextmanager
    def update_dotenv(self):  # noqa: ANN201
        """Update .env configuration.

        Yields:
            A copy of the current .env configuration dict.

        Raises:
            StoreNotSupportedError: if the project is in read-only mode.
        """
        try:
            with self.project.dotenv_update() as dotenv_file:
                yield dotenv_file
        except ProjectReadonly as err:
            raise StoreNotSupportedError(err) from err

        self._env = None


class MeltanoYmlStoreManager(SettingsStoreManager):
    """Meltano.yml Store Manager."""

    label = "`meltano.yml`"
    writable = True

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialise MeltanoYmlStoreManager instance.

        Args:
            args: Positional arguments to pass to parent class.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(*args, **kwargs)
        self._flat_config = None

    def ensure_supported(self, method: str = "get") -> None:
        """Ensure named method is supported.

        Checks that the project is not read-only and an environment is active.

        Args:
            method: Setting method (get, set, etc.)

        Raises:
            StoreNotSupportedError: If the project is read-only or named method
                is not "get".
        """
        if method != "get" and self.project.readonly:
            raise StoreNotSupportedError(ProjectReadonly())

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,
    ) -> tuple[str, dict]:
        """Get value by name from the system database.

        Args:
            name: Setting name.
            setting_def: SettingDefinition.
            cast_value: Whether to cast the value according to `setting_def`.

        Returns:
            A tuple the got value and a dictionary containing metadata.

        Raises:
            ConflictingSettingValueException: if multiple conflicting values
                for the same setting are provided.
        """
        keys = [setting_def.name, *setting_def.aliases] if setting_def else [name]
        flat_config = self.flat_config
        vals_with_metadata = []
        for key in keys:
            try:
                value = flat_config[key]
            except KeyError:
                continue

            self.log(f"Read key '{key}' from `meltano.yml`: {value!r}")
            vals_with_metadata.append((value, {"key": key, "expandable": True}))

        if len(vals_with_metadata) > 1 and not reduce(
            eq,
            (val for val, _ in vals_with_metadata),
        ):
            raise ConflictingSettingValueException(
                metadata["key"] for _, metadata in vals_with_metadata
            )

        value, metadata = vals_with_metadata[0] if vals_with_metadata else (None, {})
        return (
            cast_setting_value(value, metadata, setting_def)
            if cast_value
            else (value, metadata)
        )

    def set(
        self,
        name: str,
        path: list[str],
        value: t.Any,  # noqa: ANN401
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Set value by name in the Meltano YAML File.

        Args:
            name: Setting name.
            path: Setting path.
            value: New value to set.
            setting_def: SettingDefinition.

        Returns:
            An empty dictionary.
        """
        keys_to_unset = (
            [setting_def.name, *setting_def.aliases] if setting_def else [name]
        )
        paths_to_unset = [key for key in keys_to_unset if "." in key]

        if len(path) == 1:
            # No need to unset `name`, since it will be overridden anyway
            keys_to_unset.remove(name)
        elif name.split(".") == path:
            # No need to unset `name` as path, since it will be overridden anyway
            paths_to_unset.remove(name)

        with self.update_config() as config:
            for key in keys_to_unset:
                config.pop(key, None)
                self.log(f"Popped key '{key}' in `meltano.yml`")

            for path_to_unset in paths_to_unset:
                pop_at_path(config, path_to_unset, None)
                self.log(f"Popped path '{path_to_unset}' in `meltano.yml`")

            set_at_path(config, path, value)
            self.log(f"Set path '{path}' in `meltano.yml`: {value!r}")

        return {}

    def unset(
        self,
        name: str,
        path: list[str],
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Unset value by name in the Meltano YAML file.

        Args:
            name: Setting name.
            path: Setting path.
            setting_def: SettingDefinition instance.

        Returns:
            An empty dictionary.
        """
        keys_to_unset = [name]
        if setting_def:
            keys_to_unset = [setting_def.name, *setting_def.aliases]

        paths_to_unset = [key for key in keys_to_unset if "." in key]

        with self.update_config() as config:
            for key in keys_to_unset:
                config.pop(key, None)
                self.log(f"Popped key '{key}' in `meltano.yml`")

            for path_to_unset in paths_to_unset:
                pop_at_path(config, path_to_unset, None)
                self.log(f"Popped path '{path_to_unset}' in `meltano.yml`")

            pop_at_path(config, path, None)
            self.log(f"Popped path '{path}' in `meltano.yml`")

        return {}

    def reset(self) -> dict:
        """Reset all Setting values in this store.

        Returns:
            An empty dictionary.
        """
        with self.update_config() as config:
            config.clear()
        return {}

    @property
    def flat_config(self) -> dict:
        """Get dictionary of flattened configuration.

        Returns:
            A dictionary of flattened configuration.
        """
        if self._flat_config is None:
            self._flat_config = self.settings_service.flat_meltano_yml_config
        return self._flat_config

    @contextmanager
    def update_config(self):  # noqa: ANN201
        """Update Meltano YAML configuration.

        Yields:
            A copy of the current YAML configuration dict.

        Raises:
            StoreNotSupportedError: if the project is in read-only mode.
        """
        config = deepcopy(self.settings_service.meltano_yml_config)
        yield config

        try:
            self.settings_service.update_meltano_yml_config(config)
        except ProjectReadonly as err:
            raise StoreNotSupportedError(err) from err

        self._flat_config = None

        # This is not quite the right place for this, but we need to create
        # setting defs for missing keys again when `meltano.yml` changes
        self.settings_service._setting_defs = None


class MeltanoEnvStoreManager(MeltanoYmlStoreManager):
    """Configuration stored in an environment within `meltano.yml`."""

    label = "the active environment in `meltano.yml`"

    def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Initialise MeltanoEnvStoreManager instance.

        Args:
            args: Positional arguments to pass to parent class.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(*args, **kwargs)
        self.ensure_supported()

    @property
    def flat_config(self) -> dict[str, t.Any]:
        """Get dictionary of flattened configuration.

        Returns:
            A dictionary of flattened configuration.
        """
        if self._flat_config is None:
            self._flat_config = flatten(self.settings_service.environment_config, "dot")
        return self._flat_config

    def ensure_supported(self, method: str = "get") -> None:
        """Ensure project is not read-only and an environment is active.

        Args:
            method: Setting method (get, set, etc.)

        Raises:
            StoreNotSupportedError: if the project is read-only or no
                environment is active.
        """
        super().ensure_supported(method)
        if not self.settings_service.supports_environments:
            reason = "Project config cannot be stored in an Environment"
            raise StoreNotSupportedError(reason)
        if self.settings_service.project.environment is None:
            raise StoreNotSupportedError(NoActiveEnvironment())

    @contextmanager
    def update_config(self) -> None:
        """Update Meltano Environment configuration.

        Yields:
            A copy of the current environment configuration dict.

        Raises:
            StoreNotSupportedError: if the project is in read-only mode.
        """
        config = deepcopy(self.settings_service.environment_config)
        yield config

        try:
            self.settings_service.update_meltano_environment_config(config)
        except ProjectReadonly as err:
            raise StoreNotSupportedError(err) from err

        self._flat_config = None

        # This is not quite the right place for this, but we need to create
        # setting defs for missing keys again when `meltano.yml` changes
        self.settings_service._setting_defs = None


class DbStoreManager(SettingsStoreManager):
    """Store manager for settings stored in the system database."""

    label = "the system database"
    writable = True

    def __init__(
        self,
        *args,  # noqa: ANN002
        bulk: bool = False,
        session: Session | None = None,
        **kwargs,  # noqa: ANN003
    ):
        """Initialise DbStoreManager.

        Args:
            args: Positional arguments to pass to parent class.
            bulk: Flag to determine whether parent metadata is returned alongside child.
            session: SQLAlchemy Session to use when querying the system database.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(*args, **kwargs)
        self.session = session
        self.ensure_supported()
        self.bulk = bulk
        self._all_settings = None

    def ensure_supported(
        self,
        method: str = "get",  # noqa: ARG002
    ) -> None:
        """Return True if passed method is supported by this store.

        Args:
            method: Unused. Included to match parent class method signature.

        Raises:
            StoreNotSupportedError: if database session is not provided.
        """
        if not self.session:
            raise StoreNotSupportedError("No database session provided")  # noqa: EM101

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,  # noqa: ARG002
        *,
        cast_value: bool = False,  # noqa: ARG002
    ) -> tuple[str, dict]:
        """Get value by name from the system database.

        Args:
            name: Setting name.
            setting_def: Unused. Included to match parent class method signature.
            cast_value: Whether to cast the value according to `setting_def`.

        Returns:
            A tuple the got value and an empty dictionary.
        """
        try:
            if self.bulk:
                value = self.all_settings[name]
            else:
                value = (
                    self.session.query(Setting)
                    .filter_by(namespace=self.namespace, name=name, enabled=True)
                    .one()
                    .value
                )

            self.log(f"Read key '{name}' from system database: {value!r}")
            return value, {}
        except (sqlalchemy.orm.exc.NoResultFound, KeyError):
            return None, {}

    def set(
        self,
        name: str,
        path: list[str],  # noqa: ARG002
        value: t.Any,  # noqa: ANN401
        setting_def: SettingDefinition | None = None,  # noqa: ARG002
    ) -> dict:
        """Set value by name in the system database.

        Args:
            name: Setting name.
            path: Unused. Included to match parent class method signature.
            value: New value to set.
            setting_def: Unused. Included to match parent class method signature.

        Returns:
            An empty dictionary.
        """
        setting = Setting(
            namespace=self.namespace,
            name=name,
            value=value,
            enabled=True,
        )
        self.session.merge(setting)
        self.session.commit()

        self._all_settings = None

        self.log(f"Set key '{name}' in system database: {value!r}")
        return {}

    def unset(
        self,
        name: str,
        path: list[str],  # noqa: ARG002
        setting_def: SettingDefinition | None = None,  # noqa: ARG002
    ) -> dict:
        """Unset value by name in the system database store.

        Args:
            name: Setting name.
            path: Unused. Included to match parent class method signature.
            setting_def: Unused. Included to match parent class method signature.

        Returns:
            An empty dictionary.
        """
        self.session.query(Setting).filter_by(
            namespace=self.namespace,
            name=name,
        ).delete()
        self.session.commit()

        self._all_settings = None

        self.log(f"Deleted key '{name}' from system database")
        return {}

    def reset(self) -> dict:
        """Reset all Setting values in this store.

        Returns:
            An empty dictionary.
        """
        self.session.query(Setting).filter_by(namespace=self.namespace).delete()
        self.session.commit()

        self._all_settings = None

        return {}

    @property
    def namespace(self) -> str:
        """Return the current SettingService namespace.

        Returns:
            The current SettingService namespace
        """
        return self.settings_service.db_namespace

    @property
    def all_settings(self) -> dict[str, Setting]:
        """Fetch all settings from the system database for this namespace that are enabled.

        Returns:
            A dictionary of Setting models.
        """  # noqa: E501
        if self._all_settings is None:
            self._all_settings = {
                setting.name: setting.value
                for setting in self.session.query(Setting)
                .filter_by(namespace=self.namespace, enabled=True)
                .all()
            }
        return self._all_settings


class InheritedStoreManager(SettingsStoreManager):
    """Store manager for settings inherited from a parent plugin."""

    label = "inherited"

    def __init__(
        self,
        settings_service: SettingsService,
        *args,  # noqa: ANN002
        bulk: bool = False,
        **kwargs,  # noqa: ANN003
    ):
        """Initialize inherited store manager.

        Args:
            settings_service: SettingsService instance.
            args: Positional arguments to pass to parent class.
            bulk: Flag to determine whether parent metadata is returned alongside child.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(settings_service, *args, **kwargs)
        self._kwargs = {**kwargs, "expand_env_vars": False}
        self.bulk = bulk
        self._config_with_metadata = None

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,  # noqa: ARG002
    ) -> tuple[str, dict]:
        """Get a Setting value by name and SettingDefinition.

        Args:
            name: Setting name.
            setting_def: SettingDefinition.
            cast_value: Whether to cast the value according to `setting_def`.

        Returns:
            A tuple containing the got value and accompanying metadata dictionary.

        Raises:
            StoreNotSupportedError: if no setting_def is passed.
        """
        if not setting_def:
            raise StoreNotSupportedError("Setting definition is missing")  # noqa: EM101

        if not self.inherited_settings_service:
            raise StoreNotSupportedError("Inherited settings service is missing")  # noqa: EM101

        value, metadata = self.get_with_metadata(setting_def.name)
        if value is None or metadata["source"] is SettingValueStore.DEFAULT:
            return None, {}

        self.log(f"Read key '{name}' from inherited: {value!r}")
        return value, {
            "inherited_source": metadata["source"],
            "expandable": metadata.get("expandable", False),
        }

    @property
    def inherited_settings_service(self) -> SettingsService:
        """Return settings service to inherit configuration from.

        Returns:
            A SettingsService to inherit configuration from.
        """
        return self.settings_service.inherited_settings_service

    @property
    def config_with_metadata(self) -> dict:
        """Return all inherited config and metadata.

        Returns:
            A dictionary containing config and metadata.
        """
        if self._config_with_metadata is None:
            self._config_with_metadata = (
                self.inherited_settings_service.config_with_metadata(**self._kwargs)
            )
        return self._config_with_metadata

    def get_with_metadata(self, name: str) -> tuple[str, dict]:
        """Return inherited config and metadata for the named setting.

        Args:
            name: Setting name.

        Returns:
            A tuple containing the got value and accompanying metadata dictionary.
        """
        if self.bulk:
            metadata = self.config_with_metadata[name]
            return metadata["value"], metadata

        return self.inherited_settings_service.get_with_metadata(name, **self._kwargs)


class DefaultStoreManager(SettingsStoreManager):
    """Store manager for defaults supplied by the SettingDefinition."""

    label = "the default"

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,  # noqa: ARG002
    ) -> tuple[str, dict]:
        """Get a Setting value by name and SettingDefinition.

        Args:
            name: Setting name.
            setting_def: SettingDefinition.
            cast_value: Whether to cast the value according to `setting_def`.

        Returns:
            A tuple containing the got value and accompanying metadata dictionary.
        """
        value = None
        metadata = {}
        if setting_def:
            value = setting_def.value
            if value is not None:
                self.log(f"Read key '{name}' from default: {value!r}")
                return value, {"expandable": True}
        # As default is lowest in our order of precedence, we want it to always return
        # a value, even if it is None.
        return value, metadata


class AutoStoreManager(SettingsStoreManager):
    """Manager that determines the appropriate store based on current context."""

    label = "the system database, `meltano.yml`, and `.env`"
    writable = True

    def __init__(self, *args, cache: bool = True, **kwargs):  # noqa: ANN002, ANN003
        """Initialise AutoStoreManager.

        Args:
            args: Positional arguments to pass to parent class.
            cache: Cache store manager instances for fast retrieval.
            kwargs: Keyword arguments to pass to parent class.
        """
        super().__init__(*args, **kwargs)
        self.cache = cache
        self._kwargs = {"settings_service": self.settings_service, **kwargs}
        self._managers = {}

    def manager_for(self, store: SettingValueStore) -> SettingsStoreManager:
        """Get setting store manager for this a value store.

        Args:
            store: A setting value store.

        Returns:
            Setting store manager.
        """
        if not self.cache or store not in self._managers:
            self._managers[store] = store.manager(**self._kwargs)
        return self._managers[store]

    @property
    def sources(self) -> list[SettingValueStore]:
        """Return a list of readable SettingValueStore.

        Returns:
            A list of readable SettingValueStore
        """
        sources = SettingValueStore.readables()
        sources.remove(SettingValueStore.AUTO)
        return sources

    @property
    def stores(self) -> list[SettingValueStore]:
        """Return a list of writable SettingValueStore.

        Returns:
            A list of writable SettingValueStore
        """
        stores = SettingValueStore.writables()
        stores.remove(SettingValueStore.AUTO)
        return stores

    def ensure_supported(self, store, method="set") -> bool | None:  # noqa: ANN001
        """Return if a given store is supported for the given method.

        Args:
            store: The store to check.
            method: The method to check for the given store.

        Returns:
            True if store supports method.
        """
        try:
            manager = self.manager_for(store)
            manager.ensure_supported(method)
            return True
        except StoreNotSupportedError:
            return False

    def auto_store(  # Too complex
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
    ) -> SettingValueStore | None:
        """Get first valid writable SettingValueStore instance for a Setting.

        Args:
            name: Setting name.
            setting_def: The setting definition. If `None`, one will be
                discovered using `self.find_setting(name)`.

        Returns:
            A SettingValueStore, if found, else None.
        """
        setting_def = setting_def or self.find_setting(name)

        # only the system database is available in readonly mode
        if self.project.readonly:
            if self.ensure_supported(store=SettingValueStore.DB):
                return SettingValueStore.DB
            return None

        # value is a secret
        if setting_def and setting_def.is_redacted:
            if self.ensure_supported(store=SettingValueStore.DOTENV):
                return SettingValueStore.DOTENV
            if self.ensure_supported(store=SettingValueStore.DB):
                return SettingValueStore.DB
            # ensure secrets don't leak into other stores
            return None

        # value is env-specific
        if (
            setting_def
            and setting_def.env_specific
            and self.ensure_supported(store=SettingValueStore.DOTENV)
        ):
            return SettingValueStore.DOTENV

        # no active meltano environment
        if not self.project.environment:
            # return root `meltano.yml`
            if self.ensure_supported(store=SettingValueStore.MELTANO_YML):
                return SettingValueStore.MELTANO_YML
            # fall back to dotenv
            if self.ensure_supported(store=SettingValueStore.DOTENV):
                return SettingValueStore.DOTENV
            # fall back to meltano system db
            if self.ensure_supported(store=SettingValueStore.DB):
                return SettingValueStore.DB
            return None

        # any remaining config routed to meltano environment
        if self.ensure_supported(store=SettingValueStore.MELTANO_ENV):
            return SettingValueStore.MELTANO_ENV
        # Fall back to root `meltano.yml`. This is required for Meltano
        # settings, which cannot be stored in an Environment
        if self.ensure_supported(store=SettingValueStore.MELTANO_YML):
            return SettingValueStore.MELTANO_YML
        # fall back to dotenv
        if self.ensure_supported(store=SettingValueStore.DOTENV):
            return SettingValueStore.DOTENV
        # fall back to meltano system db
        if self.ensure_supported(store=SettingValueStore.DB):
            return SettingValueStore.DB
        return None

    def get(
        self,
        name: str,
        setting_def: SettingDefinition | None = None,
        *,
        cast_value: bool = False,
        **kwargs,  # noqa: ANN003
    ) -> tuple[str, dict]:
        """Get a Setting value by name and SettingDefinition.

        Args:
            name: Setting name.
            setting_def: SettingDefinition. If none is passed, one will be
                discovered using `self.find_setting(name)`.
            cast_value: Whether to cast the value according to `setting_def`.
            kwargs: Additional keyword arguments to pass to `manager.get()`.

        Returns:
            A tuple containing the got value and accompanying metadata dictionary.
        """
        setting_def = setting_def or self.find_setting(name)

        metadata = {}
        value = None
        found_source = None

        for source in self.sources:
            try:
                manager = self.manager_for(source)
            except StoreNotSupportedError:
                continue

            try:
                value, metadata = manager.get(
                    name,
                    setting_def=setting_def,
                    cast_value=cast_value,
                    **kwargs,
                )
            except StoreNotSupportedError:
                continue

            found_source = source

            if value is not None:
                break

        metadata["source"] = found_source

        if auto_store := self.auto_store(name, setting_def=setting_def):
            metadata["auto_store"] = auto_store
            metadata["overwritable"] = auto_store.can_overwrite(found_source)

        return (
            cast_setting_value(value, metadata, setting_def)
            if cast_value
            else (value, metadata)
        )

    def set(self, name: str, path: list[str], value, setting_def=None) -> dict:  # noqa: ANN001
        """Set a Setting by name, path and (optionally) SettingDefinition.

        Args:
            name: Setting name.
            path: Setting path.
            value: New Setting value.
            setting_def: Setting definition. If none is passed, one will be
                discovered using `self.find_setting(name)`.

        Returns:
            A dictionary of metadata pertaining to the set operation.

        Raises:
            StoreNotSupportedError: exception encountered when attempting to
                write to store.
        """
        setting_def = setting_def or self.find_setting(name)
        store = self.auto_store(name, setting_def=setting_def)
        logger.debug(f"AutoStoreManager returned store '{store}'")  # noqa: G004
        if store is None:
            raise StoreNotSupportedError("No storage method available")  # noqa: EM101

        # May raise StoreNotSupportedError, but that's good.
        manager = self.manager_for(store)
        metadata = manager.set(name, path, value, setting_def=setting_def)
        metadata["store"] = store
        return metadata

    def unset(
        self,
        name: str,
        path: list[str],
        setting_def: SettingDefinition | None = None,
    ) -> dict:
        """Unset value, by name, path and `SettingDefinition`, in all stores.

        Args:
            name: Setting name.
            path: Setting path.
            setting_def: Setting definition. If none is passed, one will be
                discovered using `self.find_setting(name)`.

        Returns:
            A metadata dictionary containing details of the last value unset.

        Raises:
            error: exception encountered when trying to write to a store.
        """
        setting_def = setting_def or self.find_setting(name)

        error = None
        metadata = {}

        for store in self.stores:
            try:
                manager = self.manager_for(store)
                value, _ = manager.get(name, setting_def=setting_def)
            except StoreNotSupportedError:
                continue

            try:
                metadata = manager.unset(name, path, setting_def=setting_def)
                metadata["store"] = store
            except StoreNotSupportedError as err:
                # Only raise if we're sure we were going to unset something
                if value is not None:
                    error = err

        if error:
            raise error

        return metadata

    def reset(self) -> dict:
        """Reset all stores.

        Returns:
            An empty dictionary.
        """
        for store in self.stores:
            with suppress(StoreNotSupportedError):
                manager = self.manager_for(store)
                manager.reset()
        return {}

    def find_setting(self, name: str) -> SettingDefinition | None:
        """Find a setting by name.

        Args:
            name: Name of the Setting to find.

        Returns:
            A matching SettingDefinition, if found, else None.
        """
        try:
            return self.settings_service.find_setting(name)
        except SettingMissingError:
            return None
